﻿/*
 * © EUROPEAN COMMISSION, 2009-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la
«Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que
conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse
suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le
logiciel distribué sous la Licence est distribué «en
l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT,
expresses ou implicites.
 * Consultez la Licence pour les autorisations et les
restrictions linguistiques spécifiques relevant de la
Licence.
 *
 * Auteur de l'oeuvre: Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Configuration;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.IO.Pipes;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Soap;
using System.Runtime.Serialization.Formatters.Binary;
using MySql.Data.MySqlClient;

namespace mDownloaderService
{
	public partial class mDownloaderService : ServiceBase
	{
		/// <summary>
		/// Object interfacing with the database
		/// </summary>
		public class DownloadDatabase : IDisposable
		{
			/// <summary>
			/// Locker for the DataReader
			/// </summary>
			private object lockDataReader;

			/// <summary>
			/// MySQL connection
			/// </summary>
			private MySqlConnection db;
			/// <summary>
			/// Encoding used when writing to the database
			/// </summary>
			private Encoding dbEncoding;

			/// <summary>
			/// Build a DownloadDatabase object
			/// </summary>
			/// <param name="connectionString"></param>
			/// <param name="encoding"></param>
			public DownloadDatabase( string connectionString, Encoding encoding )
			{
				lockDataReader = new object();
				db = new MySqlConnection(connectionString);
				dbEncoding = encoding;
			}

			/// <summary>
			/// Load requests from the database
			/// </summary>
			/// <param name="CallId"></param>
			/// <returns></returns>
			public ClientRequest[] FetchRequests(long CallId)
			{
				if (!db.Ping())
				{
					db.Open();
				}

				List<ClientRequest> requestList = new List<ClientRequest>();
				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "select `url_id`, `url`, `method`, `use_proxy`, `post_fields` from `ajax_call_urls` where `call_id` = @call_id";
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@call_id", CallId);
					query.Prepare();

					lock (lockDataReader)
					{
						using (MySqlDataReader reader = query.ExecuteReader())
						{
							if (reader.HasRows)
							{
								while (reader.Read())
								{
									string postData = string.Empty;
									if (!reader.IsDBNull(4))
									{
										postData = reader.GetString(4);
									}
									DownloadRequest dlReq = new DownloadRequest(
										reader.GetString("url"),
										reader.GetString("use_proxy").ToLower() == "yes" ? true : false,
										reader.GetString("method").ToLower() == "post" ? DownloadRequest.HttpMethod.post : DownloadRequest.HttpMethod.get,
										Encoding.Convert(Encoding.Unicode, Encoding.UTF8, Encoding.Unicode.GetBytes(postData))
									);
									ClientRequest clReq = new ClientRequest(
										CallId,
										reader.GetInt64("url_id"),
										dlReq
									);
									requestList.Add(clReq);
								}
							}
						}
					}
				}
				return requestList.ToArray();
			}

			/// <summary>
			/// Build ClientRequest from DownloadRequest
			/// Automatically fetch a CallID and an UrlID from the database
			/// </summary>
			/// <param name="request"></param>
			/// <returns></returns>
			public ClientRequest DownloadRequestToClientRequest(DownloadRequest request)
			{
				ClientRequest[] resp = DownloadRequestToClientRequest(new DownloadRequest[] { request });
				if (null != resp && resp.Length == 1)
				{
					return resp[0];
				}
				return null;
			}
			/// <summary>
			/// Build ClientRequests from DownloadRequests
			/// Automatically fetch a CallID and UrlIDs from the database for several DownloadRequests (they will all share the same CallId, but will have different UrlId)
			/// </summary>
			/// <param name="request"></param>
			/// <returns></returns>
			public ClientRequest[] DownloadRequestToClientRequest(DownloadRequest[] requests)
			{
				if (0 == requests.Length)
				{
					return null;
				}
				if (!db.Ping())
				{
					db.Open();
				}

				List<ClientRequest> requestList = new List<ClientRequest>();

				long callId = 0;
				int affectedRows = 0;

				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "insert into `ajax_calls` (`call_time`, `user_id`) values(now(), NULL) ";
					query.Prepare();
					lock (lockDataReader)
					{
						affectedRows = query.ExecuteNonQuery();
						callId = query.LastInsertedId;
					}
					
				}
				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "insert into `ajax_call_urls` (`call_id`, `url_id`, `url`, `method`, `use_proxy`, `post_fields`) values(@call_id, @url_id, @url, @method, @use_proxy, @data) ";
					long urlId = 0;
					foreach (DownloadRequest dlr in requests)
					{
						if (!query.IsPrepared)
						{
							query.Parameters.AddWithValue("@call_id", callId);
							query.Parameters.AddWithValue("@url_id", urlId);
							query.Parameters.AddWithValue("@url", dlr.url);
							query.Parameters.AddWithValue("@method", dlr.method == DownloadRequest.HttpMethod.post ? "post" : "get");
							query.Parameters.AddWithValue("@use_proxy", dlr.throughProxy ? "yes" : "no");
							if(DownloadRequest.HttpMethod.post == dlr.method && null != dlr.data) {
								query.Parameters.AddWithValue("@data", dlr.data);
							} else {
								query.Parameters.AddWithValue("@data", DBNull.Value);
							}
							query.Prepare();
						}
						else
						{
							query.Parameters[0].Value = callId;
							query.Parameters[1].Value = urlId;
							query.Parameters[2].Value = dlr.url;
							query.Parameters[3].Value = dlr.method == DownloadRequest.HttpMethod.post ? "post" : "get";
							query.Parameters[4].Value = dlr.throughProxy ? "yes" : "no";
							if (DownloadRequest.HttpMethod.post == dlr.method && null != dlr.data)
							{
								query.Parameters[5].Value = dlr.data;
							}
							else
							{
								query.Parameters[5].Value = DBNull.Value;
							}
						}
						lock (lockDataReader)
						{
							affectedRows = query.ExecuteNonQuery();
						}
						requestList.Add(new ClientRequest((int)callId, urlId, dlr));
						urlId++;
					}
				}
				return requestList.ToArray();
			}

			/// <summary>
			/// Save the downloaded document (or error information) in the database
			/// </summary>
			/// <param name="request"></param>
			/// <param name="response"></param>
			/// <returns></returns>
			public int SaveRequestResponse( ClientRequest request, mWebDocument response )
			{
				if (!db.Ping())
				{
					db.Open();
				}
				
				int affectedRows = 0;
				using (MySqlCommand query = db.CreateCommand())
				{
					// If not binary, convert document encoding to match database's internal encoding (default: utf-8)
					byte[] documentContent = null;
					if (null != response.LastException)
					{
						documentContent = new byte[0];
					}
					else
					{
						if (response.IsTextContentType && response.HasSupportedCharacterSet)
						{
							documentContent = Encoding.Convert(Encoding.Unicode, dbEncoding, response.Body);
						}
						else
						{
							documentContent = response.Body;
						}
					}

					query.CommandType = CommandType.Text;
					query.CommandText =
						"update `ajax_call_urls` set " + Environment.NewLine +
						"`url`    = @url," + Environment.NewLine +
						"`http_code`    = @http_code," + Environment.NewLine +
						"`call_content` = @call_content," + Environment.NewLine +
						"`call_content_length` = @call_content_length," + Environment.NewLine +
						"`call_content_type` = @call_content_type," + Environment.NewLine +
						"`call_header`  = @call_header," + Environment.NewLine +
						"`call_error`   = @call_error," + Environment.NewLine +
						"`duration`     = @duration" + Environment.NewLine +
						"where `call_id` = @call_id" + Environment.NewLine +
						"and `url_id` = @url_id" + Environment.NewLine;

					query.Parameters.AddWithValue("@call_id", request.callId);
					query.Parameters.AddWithValue("@url_id", request.urlId);
					query.Parameters.AddWithValue("@url", response.ResponseUri);
					query.Parameters.AddWithValue("@http_code", (int)response.StatusCode);
					query.Parameters.AddWithValue("@call_content", documentContent);
					query.Parameters.AddWithValue("@call_content_length", (null == documentContent || documentContent.Length == 0) ? 0 : documentContent.Length);
					query.Parameters.AddWithValue("@call_content_type", response.ContentType);
					StringBuilder strHeaders = new StringBuilder();
					if (null != response.Headers)
					{
						foreach (string key in response.Headers.AllKeys)
						{
							strHeaders.AppendLine(key + ": " + response.Headers.Get(key));
						}
					}
					query.Parameters.AddWithValue("@call_header", strHeaders.ToString());
					query.Parameters.AddWithValue("@duration", response.Duration.TotalSeconds);
					if (!response.Success)
					{
						string strCallError = response.LastException.Message + Environment.NewLine;
						strCallError += response.LastException.StackTrace;
						query.Parameters.AddWithValue("@call_error", strCallError);
					}
					else
					{
						query.Parameters.AddWithValue("@call_error", null);
					}
					query.Prepare();

					lock (lockDataReader)
					{
						affectedRows = query.ExecuteNonQuery();
					}
				}
				return affectedRows;
			}

			/// <summary>
			/// Fetches the number of pending download requests for a given Call Id
			/// </summary>
			/// <param name="CallId"></param>
			/// <returns>The number of requests still being processed</returns>
			public long FetchPendingDownloads(long CallId)
			{
				if (!db.Ping())
				{
					db.Open();
				}

				long pending = 0;
				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "select count(*) as `pending` " +
							"from `ajax_call_urls` " +
							"where (`call_error` is null and `http_code` is null) "+
							"and `call_id` = @call_id ";
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@call_id", CallId);
					query.Prepare();

					lock (lockDataReader)
					{
						using (MySqlDataReader reader = query.ExecuteReader())
						{
							if (reader.HasRows)
							{
								reader.Read();
								pending = reader.GetInt64("pending");
							}
						}
					}
				}
				return pending;
			}

			/// <summary>
			/// Fetch some documents of a given call from the database
			/// </summary>
			/// <param name="CallId"></param>
			/// <param name="UrlIdList">List of UrlId of the documents to fetch (from 0 to the number of reqeusts -1), or null to fetech all documents</param>
			/// <returns>The Call's results</returns>
			public SavedCallResult FetchResults(long CallId, int[] UrlIdList)
			{
				if (!db.Ping())
				{
					db.Open();
				}

				SavedCallResult result = new SavedCallResult( CallId );
				List<SavedDocument> documents = new List<SavedDocument>();

				using (MySqlCommand query = db.CreateCommand())
				{
					string limitById = string.Empty;
					if (null != UrlIdList && 0 != UrlIdList.Length)
					{
						List<string> ids = new List<string>();
						foreach (int urlId in UrlIdList)
						{
							ids.Add(urlId.ToString());
						}
						limitById = "and `url_id` in (" + string.Join(", ", ids.ToArray()) + ")";
					}
					query.CommandType = CommandType.Text;
					query.CommandText = 
						"select call_id, url_id, url, post_fields, http_code, "+
						"call_content, call_content_length, call_content_type, call_header, call_error, "+
						"duration, last_update, method, use_proxy "+
						"from `ajax_call_urls` "+
						"where `call_id` = @call_id " +
						limitById;
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@call_id", CallId);
					query.Prepare();

					lock (lockDataReader)
					{
						using (MySqlDataReader reader = query.ExecuteReader())
						{
							if (reader.HasRows)
							{
								while (reader.Read())
								{
									SavedDocument document = new SavedDocument();
									document.callId = reader.GetInt64("call_id");
									document.urlId = reader.GetInt64("url_id");
									string postData = string.Empty;
									if (!reader.IsDBNull(3))
									{
										
										postData = reader.GetString(3);
										//postData = Encoding.UTF8.GetString(Encoding.UTF8.GetBytes(reader.GetString(3)));
									}
									document.request = new DownloadRequest(
										reader.GetString("url"),
										reader.GetString("use_proxy").ToLower() == "yes" ? true : false,
										reader.GetString("method").ToLower() == "post" ? DownloadRequest.HttpMethod.post : DownloadRequest.HttpMethod.get,
										//Encoding.Unicode.GetBytes(postData)
										Encoding.UTF8.GetBytes(postData)
									);
									document.response = new DownloadResponse(reader);
									documents.Add(document);
								}
							}
						}
					}
				}

				result.urls = documents.ToArray();
				return result;
			}

			/// <summary>
			/// Fetch all the documents of a given call from the database
			/// </summary>
			/// <param name="CallId"></param>
			/// <returns></returns>
			public SavedCallResult FetchResults(long CallId)
			{
				return FetchResults(CallId, null);
			}

			/// <summary>
			/// Reset responses to a request
			/// </summary>
			/// <param name="CallId"></param>
			public int ResetResponses(long CallId)
			{
				if (!db.Ping())
				{
					db.Open();
				}

				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "update `ajax_call_urls` " +
							"set " +
							"`http_code` = null, " +
							"`call_content` = null, " +
							"`call_content_length` = null, " +
							"`call_content_type` = null, " +
							"`call_header` = null, " +
							"`call_error` = null, " +
							"`duration` = null, " +
							"`last_update` = null " +
							"where `call_id` = @call_id ";
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@call_id", CallId);
					query.Prepare();

					lock (lockDataReader)
					{
						return query.ExecuteNonQuery();
					}
				}
			}

			/// <summary>
			/// Reset responses to a request
			/// </summary>
			/// <param name="CallId"></param>
			public int DeleteRequest(long CallId)
			{
				if (!db.Ping())
				{
					db.Open();
				}

				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "delete u.*, c.* " +
							"from `ajax_calls` c " +
							"  left join `ajax_call_urls` u " +
							"  on c.`call_id` = u.`call_id` " +
							"where c.`call_id` = @call_id ";
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@call_id", CallId);
					query.Prepare();

					lock (lockDataReader)
					{
						return query.ExecuteNonQuery();
					}
				}
			}





			~DownloadDatabase()
			{
				Dispose(false);
			}
			private bool disposed = false;
			public void Dispose(bool disposing)
			{
				if (!disposed)
				{
					if (disposing)
					{
						// It is safe to access other objects here
						db.Dispose();
					}
					disposed = true;
				}
			}
			public void Dispose()
			{
				Dispose(true);
				GC.SuppressFinalize(this);
			}

			public bool isDisposed
			{
				get
				{
					return disposed;
				}
			}
		}
	}
}