﻿/*
 * mTools - © EUROPEAN COMMISSION, 2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous
 * la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou 
 * implicites.
 * Consultez la Licence pour les autorisations et les restrictions 
 * linguistiques spécifiques relevant de la Licence.
 *
 * Auteur de l'oeuvre: Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Net;
using System.Net.Sockets;
using System.IO;

using System.Runtime.Serialization;

namespace MetaNull.Net
{
	/// <summary>
	/// Holds a resource downlaoded from the Web
	/// </summary>
	[Serializable()]
	public class WebDocument
	{
		/// <summary>
		/// Get the content of the exception in case of an error
		/// </summary>
		public Exception LastException { get { return _LastException;  } }
		private Exception _LastException;

		/// <summary>
		/// Indicate if the operation was successfull or not
		/// </summary>
		public bool Success { get { return null == _LastException; } }

		/// <summary>
		/// Get the method that is used to encode the body of the response
		/// </summary>
		public string ContentEncoding { get { return _ContentEncoding; } }
		private string _ContentEncoding;

		/// <summary>
		/// Get the Character Set of the response
		/// </summary>
		public Encoding CharacterSet{get {return _CharacterSet;} }
		private Encoding _CharacterSet;

		/// <summary>
		/// Get the name of the Character set of the response
		/// </summary>
		public string CharacterSetName {get{return _CharacterSetName; }}
		private string _CharacterSetName;

		/// <summary>
		/// Check if the response provides Character Set information
		/// </summary>
		public bool HasCharacterSet { get { return string.IsNullOrEmpty(_CharacterSetName); } }

		/// <summary>
		/// Check if the response provides a Character Set supported by the local system
		/// </summary>
		public bool HasSupportedCharacterSet { get { return null != _CharacterSet; }}
		
		/// <summary>
		/// Indicates if the response's content was automatically converted to unicode
		/// </summary>
		public bool WasCharacterSetConverted { get {return _WasCharacterSetConverted; }}
		private bool _WasCharacterSetConverted = false;

		/// <summary>
		/// Get the length of the content returned by the request
		/// </summary>
		public long ContentLength { get { return _ContentLength; } }
		private long _ContentLength;

		/// <summary>
		/// Get the length of the content returned by the request
		/// </summary>
		public string ContentType{get{return _ContentType;}}
		private string _ContentType;

		/// <summary>
		/// Check if the Content-Type represents a textual type of data (text/plain, text/xml, text/html, ...)
		/// </summary>
		/// <remarks>If it no Content-Type is set, this function will return FALSE</remarks>
		public bool IsTextContentType
		{
			get
			{
				if( string.IsNullOrEmpty(_ContentType) ) {
					return false;
				}
				if (_ContentType.StartsWith("text/"))
				{
					return true;
				}
				switch (_ContentType.ToLower())
				{
					case "application/rss+xml":
					case "application/rdf+xml":
					case "application/atom+xml":
					case "application/xml":
						{
							return true;
						}
				}
				return false;
			}
		}
		/// <summary>
		/// Check if the Content-Type header was set
		/// </summary>
		public bool HasContentType
		{
			get
			{
				if (string.IsNullOrEmpty(_ContentType))
				{
					return false;
				}
				return true;
			}
		}

		/// <summary>
		/// Get the length of the content returned by the request
		/// </summary>
		public CookieCollection Cookies { get { return _Cookies; } }
		private CookieCollection _Cookies;

		/// <summary>
		/// Get the headers that are associated with this response from the server
		/// </summary>
		public WebHeaderCollection Headers { get { return _Headers; } }
		private WebHeaderCollection _Headers;

		/// <summary>
		/// Get the content returned by the request (UNICODE encoded)
		/// </summary>
		/// <remarks>
		/// Encoding is automatically converted to Unicode for text documents
		/// </remarks>
		public byte[] ContentBytes { get { return document; } }
		/// <summary>
		/// Get the content encoded using UTF-8
		/// </summary>
		public byte[] ContentBytes_Utf8Encoded { get { return Encoding.Convert(Encoding.Unicode, Encoding.UTF8, document); } }
		/// <summary>
		/// Get the content (assuming it was actually UTF8 encoded);
		/// </summary>
		public byte[] ContentBytes_FixedFromUtf8 { get { return Encoding.Unicode.GetBytes(ContentString_FixedFromUTF8); } }
		/// <summary>
		/// Get the content encoded using ASCII
		/// </summary>
		public byte[] ContentBytes_ASCIIEncoded { get { return Encoding.Convert(Encoding.Unicode, Encoding.ASCII, document); } }
		/// <summary>
		/// Get the content (assuming it was actually ASCII encoded);
		/// </summary>
		public byte[] ContentBytes_FixedFromASCII { get { return Encoding.Unicode.GetBytes(ContentString_FixedFromASCII); } }
		/// <summary>
		/// Get the content as a string
		/// </summary>
		public string ContentString { get { if (null == document) return null; return Encoding.Unicode.GetString(document); } }
		/// <summary>
		/// Get the content (assuming it was actually ASCII encoded)
		/// </summary>
		public string ContentString_FixedFromASCII { get { if (null == document) return null; return Encoding.ASCII.GetString(document); } }
		/// <summary>
		/// Get the content (assuming it was actually UTF8 encoded)
		/// </summary>
		public string ContentString_FixedFromUTF8 { get { if (null == document) return null; return Encoding.UTF8.GetString(document); } }

		//public byte[] ContentBytes { get { return Encoding.Convert(Encoding.Unicode, Encoding.UTF8, document); } }

		/// <summary>
		/// Returns a String representation of the object
		/// </summary>
		/// <returns>The document's contetn as string</returns>
		public override string  ToString() {
			return ContentString;
		}

		// ContentBytes of the downloaded document
		private byte[] document;

		/// <summary>
		/// Gets the total duration of the request
		/// </summary>
		public TimeSpan Duration { get { return _Duration;  } }
		private TimeSpan _Duration;

		/*
		/// <summary>
		/// ContentBytes of the downloaded document as a string (Encoding is converted to unicode)
		/// </summary>
		public string UnicodeBody { 
			get { 
				if( null == _CharacterSet ) {
					return Encoding.Unicode.GetString( document );
				}
				return Encoding.UTF8.GetString(document);
				// return Encoding.Unicode.GetString( Encoding.Convert(_CharacterSet, Encoding.Unicode, document));
			} 
		}
		*/

		/// <summary>
		/// Gets a Boolean value that indicates whether this response was obtained from the cache.
		/// </summary>
		public bool IsFromCache { get { return _IsFromCache; } }
		private bool _IsFromCache;

		/// <summary>
		/// Gets a Boolean value that indicates whether both client and server were authenticated.
		/// </summary>
		public bool IsMutuallyAuthenticated { get { return _IsMutuallyAuthenticated; } }
		private bool _IsMutuallyAuthenticated;

		/// <summary>
		/// Gets the last date and time that the contents of the response were modified.
		/// </summary>
		public DateTime LastModified { get { return _LastModified; } }
		private DateTime _LastModified;

		/// <summary>
		/// Gets the method that is used to return the response.
		/// </summary>
		public string Method { get { return _Method; } }
		private string _Method;

		/// <summary>
		/// Gets the version of the HTTP protocol that is used in the response.
		/// </summary>
		public Version ProtocolVersion { get { return _ProtocolVersion; } }
		private Version _ProtocolVersion;

		/// <summary>
		/// Gets the URI of the Internet resource that responded to the request.
		/// </summary>
		public Uri ResponseUri { get { return _ResponseUri; } }
		private Uri _ResponseUri;

		/// <summary>
		/// Gets the name of the server that sent the response.
		/// </summary>
		public string Server { get { return _Server; } }
		private string _Server;

		/// <summary>
		/// Gets the status of the response.
		/// </summary>
		public HttpStatusCode StatusCode { get { return _StatusCode; } }
		private HttpStatusCode _StatusCode;

		/// <summary>
		/// Gets the status description returned with the response.
		/// </summary>
		public string StatusDescription { get { return _StatusDescription; } }
		private string _StatusDescription;

		// Private constructor... use httpGet() or httpPost() methods to factory an object
		private WebDocument()
		{
		}

		/// <summary>
		/// Perform a HTTP GET request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <returns></returns>
		public static WebDocument httpGet(string uri)
		{
			return httpRequest(method.Get, new Uri(uri), null, null);
		}

		/// <summary>
		/// Perform a HTTP POST request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <param name="postData"></param>
		/// <returns></returns>
		public static WebDocument httpPost(string uri, byte[] postData)
		{
			return httpRequest(method.Post, new Uri(uri), postData, null);
		}

		/// <summary>
		/// Perform a HTTP GET request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <returns></returns>
		public static WebDocument httpGet(Uri uri)
		{
			return httpRequest(method.Get, uri, null, null);
		}

		/// <summary>
		/// Perform a HTTP POST request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <param name="postData"></param>
		/// <returns></returns>
		public static WebDocument httpPost(Uri uri, byte[] postData)
		{
			return httpRequest(method.Post, uri, postData, null);
		}

		/// <summary>
		/// Perform a HTTP GET request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <param name="settings">Request settings</param>
		/// <returns></returns>
		public static WebDocument httpGet(string uri, WebDocumentRequestSettings settings)
		{
			return httpRequest(method.Get, new Uri(uri), null, settings);
		}

		/// <summary>
		/// Perform a HTTP POST request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <param name="postData"></param>
		/// <param name="settings">Request settings</param>
		/// <returns></returns>
		public static WebDocument httpPost(string uri, byte[] postData, WebDocumentRequestSettings settings)
		{
			return httpRequest(method.Post, new Uri(uri), postData, settings);
		}

		/// <summary>
		/// Perform a HTTP GET request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <param name="settings">Request settings</param>
		/// <returns></returns>
		public static WebDocument httpGet(Uri uri, WebDocumentRequestSettings settings)
		{
			return httpRequest(method.Get, uri, null, settings);
		}

		/// <summary>
		/// Perform a HTTP POST request
		/// </summary>
		/// <param name="uri">URI of the document</param>
		/// <param name="postData"></param>
		/// <param name="settings">Request settings</param>
		/// <returns></returns>
		public static WebDocument httpPost(Uri uri, byte[] postData, WebDocumentRequestSettings settings)
		{
			return httpRequest(method.Post, uri, postData, settings);
		}

		/// <summary>
		/// Http Methods
		/// </summary>
		protected enum method{ 
			/// <summary>
			/// HTTP Get
			/// </summary>
			Get,
			/// <summary>
			/// HTTP Post
			/// </summary>
			Post 
		};

		/// <summary>
		/// Perform an HTTP request
		/// </summary>
		/// <param name="httpMethod">Http method</param>
		/// <param name="uri">Uri of the resource</param>
		/// <param name="requestData">Http Request data</param>
		/// <param name="settings">Request settings</param>
		/// <returns>The downloaded resource</returns>
		protected static WebDocument httpRequest(method httpMethod, Uri uri, byte[] requestData, WebDocumentRequestSettings settings)
		{
			WebDocument doc = new WebDocument();
			HttpWebRequest request;
			HttpWebResponse response = null;
			DateTime tmStart = DateTime.Now;
			DateTime tmEndDownload;
			DateTime tmEndAll;

			doc._ResponseUri = uri;

			// Build the Request Object
			try {
				request = (HttpWebRequest)HttpWebRequest.Create(uri);
				if (null != settings)
				{
					if (null != settings.Proxy)
					{
						request.Proxy = settings.Proxy;
					}
					request.MaximumAutomaticRedirections = settings.MaximumAutomaticRedirections;
					request.AllowAutoRedirect = settings.AllowAutoRedirect;
					request.Timeout = settings.Timeout;
				}
				if (method.Post == httpMethod)
				{
					request.Method = "POST";
					request.ContentType = "application/x-www-form-urlencoded";
				}
				else
				{
					request.Method = "GET";
				}
				if (null != requestData)
				{
					request.ContentLength = requestData.Length;
					Stream strm = request.GetRequestStream();
					strm.Write(requestData, 0, requestData.Length);
					strm.Close();
				}
				
				request.CookieContainer = new CookieContainer();
			} catch( Exception e ) {
				doc._LastException = e;
				return doc;
			}

			// Perform the Request
			try
			{
				response = (HttpWebResponse)request.GetResponse();
				tmEndDownload = DateTime.Now;

				// Fetch the Response (Copy the content then dispose the Response object)
				doc._ContentEncoding = response.ContentEncoding;
				doc._ContentLength = response.ContentLength;
				doc._ContentType = response.ContentType;
				doc._Headers = response.Headers;
				doc._Cookies = response.Cookies;
				doc._Duration = tmEndDownload - tmStart;
				doc._CharacterSetName = response.CharacterSet;
				doc._Cookies = response.Cookies;
				doc._Headers = response.Headers;
				doc._IsFromCache = response.IsFromCache;
				doc._IsMutuallyAuthenticated = response.IsMutuallyAuthenticated;
				doc._LastModified = response.LastModified;
				doc._Method = response.Method;
				doc._ProtocolVersion = response.ProtocolVersion;
				doc._ResponseUri = response.ResponseUri;
				doc._Server = response.Server;
				doc._StatusCode = response.StatusCode;
				doc._StatusDescription = response.StatusDescription;
				if (!string.IsNullOrEmpty(doc._CharacterSetName))
				{
					try
					{
						doc._CharacterSet = Encoding.GetEncoding(doc._CharacterSetName);
					}
					catch (ArgumentException)
					{
						// unsupported character set
					}
				} else {
					if (!doc.HasContentType || doc.IsTextContentType ) {
						doc._CharacterSet = Encoding.UTF8;
					}
				}
				

				// Read the data
				if (doc.IsTextContentType && doc.HasSupportedCharacterSet )
				{
					// Text data, convert to unicode encoding
					Stream respStream = response.GetResponseStream();
					StreamReader respReader;
					respReader = new StreamReader(respStream, doc.CharacterSet);

					string stringDocument = respReader.ReadToEnd();
					respReader.Close();
					if (Encoding.Unicode.Equals(doc.CharacterSet))
					{
						doc._WasCharacterSetConverted = false;
						doc.document = Encoding.Unicode.GetBytes(stringDocument);
					}
					else
					{
						doc._WasCharacterSetConverted = true;
						doc.document = Encoding.Convert(doc.CharacterSet, Encoding.Unicode, doc.CharacterSet.GetBytes(stringDocument));
					}
				}
				else
				{
					// Binary data
					doc._WasCharacterSetConverted = false;
					Stream respStream = response.GetResponseStream();
					BinaryReader respReader = new BinaryReader(respStream);
					System.IO.MemoryStream ms = new System.IO.MemoryStream();
					int bufferSize = 4096;
					byte[] buffer;
					do
					{
						buffer = respReader.ReadBytes(bufferSize);
						ms.Write(buffer, 0, buffer.Length);
					} while (buffer.Length == bufferSize);
					respReader.Close();
					doc.document = ms.ToArray();
				}

				tmEndAll = DateTime.Now;
				doc._Duration = tmEndAll - tmStart;
			}
			catch (WebException e)
			{
				doc._LastException = e;
				if (WebExceptionStatus.ProtocolError == e.Status)
				{
					doc._StatusCode = ((HttpWebResponse)e.Response).StatusCode;
					doc._StatusDescription = ((HttpWebResponse)e.Response).StatusDescription;
				}
			}
			catch (Exception e)
			{
				doc._LastException = e;
			}
			finally
			{
				if (null != response)
				{
					response.Close();
					response = null;
				}
			}

			return doc;
		}
	}
}
