﻿/*
 * mAggregatorStatus - © EUROPEAN COMMISSION, 2009-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la
«Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que
conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse
suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le
logiciel distribué sous la Licence est distribué «en
l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT,
expresses ou implicites.
 * Consultez la Licence pour les autorisations et les
restrictions linguistiques spécifiques relevant de la
Licence.
 *
 * Auteur de l'oeuvre: Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using MySql.Data.MySqlClient;

namespace mAggregatorService
{
	public class FeedDatabase : IDisposable
	{
		/// <summary>
		/// Locker for the DataReader
		/// </summary>
		private object lockDataReader;

		/// <summary>
		/// MySQL connection
		/// </summary>
		private MySqlConnection db;
		/// <summary>
		/// Encoding used when writing to the database
		/// </summary>
		private Encoding dbEncoding;

		/// <summary>
		/// Build a DownloadDatabase object
		/// </summary>
		/// <param name="connectionString"></param>
		/// <param name="encoding"></param>
		public FeedDatabase(string connectionString, Encoding encoding)
		{
			lockDataReader = new object();
			db = new MySqlConnection(connectionString);
			dbEncoding = encoding;
		}

		~FeedDatabase()
		{
			Dispose(false);
		}
		private bool disposed = false;
		public void Dispose(bool disposing)
		{
			if (!disposed)
			{
				if (disposing)
				{
					// It is safe to access other objects here
					db.Dispose();
				}
				disposed = true;
			}
		}
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		public bool isDisposed
		{
			get
			{
				return disposed;
			}
		}

		/// <summary>
		/// Returns the list of feeds that need to be refreshed
		/// </summary>
		/// <returns></returns>
		public Feed[] GetOutdatedFeeds()
		{
			List<Feed> feeds = new List<Feed>();
			if (!db.Ping())
			{
				db.Open();
			}

			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText =
					"select * " +
					"from `feeds` f " +
					"where now() >= timestampadd(minute,`refresh_rate`,`last_refresh`)" +
					"or `last_refresh` is null";
				query.Prepare();


				lock (lockDataReader)
				{
					using (MySqlDataReader reader = query.ExecuteReader())
					{
						if (reader.HasRows)
						{
							while (reader.Read())
							{
								feeds.Add(new Feed(reader));
							}
						}
						reader.Close();
					}
				}
			}
			return feeds.ToArray();
		}

		/// <summary>
		/// Create a new requestedFeed
		/// </summary>
		/// <param name="Url"></param>
		/// <param name="Internet"></param>
		/// <param name="UpdateRate"></param>
		/// <returns>the Feed Id</returns>
		public long CreateFeed(Uri Url, bool Internet, int UpdateRate)
		{
			if (!db.Ping())
			{
				db.Open();
			}
			int affectedRows = 0;
			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText = "insert into `feeds` (`url`, `refresh_rate`, `use_proxy`) values (@Url, @UpdateRate, @Internet)";
				query.Parameters.AddWithValue("@Url", Url.AbsoluteUri);
				query.Parameters.AddWithValue("@UpdateRate", UpdateRate);
				query.Parameters.AddWithValue("@Internet", Internet);
				query.Prepare();
				long feedId = 0;
				lock (lockDataReader)
				{
					affectedRows = query.ExecuteNonQuery();
					feedId = query.LastInsertedId;
				}
				return feedId;
			}
		}

		/// <summary>
		/// Update the requestedFeed's last update time
		/// </summary>
		/// <param name="FeedId"></param>
		public void TouchFeed(long FeedId)
		{
			if (!db.Ping())
			{
				db.Open();
			}
			int affectedRows = 0;
			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText = "update `feeds` set `last_refresh` = now() where `feed_id` = @FeedId";
				query.Parameters.Clear();
				query.Parameters.AddWithValue("@FeedId", FeedId);
				query.Prepare();
				long feedId = 0;
				lock (lockDataReader)
				{
					affectedRows = query.ExecuteNonQuery();
					feedId = query.LastInsertedId;
				}
			}
		}

		/// <summary>
		/// Remove the requestedFeed's last update time
		/// </summary>
		/// <param name="FeedId"></param>
		public void ResetFeedTime(long FeedId)
		{
			if (!db.Ping())
			{
				db.Open();
			}
			int affectedRows = 0;
			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText = "update `feeds` set `last_refresh` = null where `feed_id` = @FeedId";
				query.Parameters.Clear();
				query.Parameters.AddWithValue("@FeedId", FeedId);
				query.Prepare();
				lock (lockDataReader)
				{
					affectedRows = query.ExecuteNonQuery();
				}
			}
		}

		/// <summary>
		/// Remove all articles from the requestedFeed
		/// </summary>
		/// <param name="FeedId"></param>
		public void TruncateFeed(long FeedId)
		{
			if (!db.Ping())
			{
				db.Open();
			}
			int affectedRows = 0;

			using (MySqlTransaction transaction = db.BeginTransaction())
			{
				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "delete from `articles` where `feed_id` = @FeedId";
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@FeedId", FeedId);
					query.Transaction = transaction;
					query.Prepare();
					lock (lockDataReader)
					{
						affectedRows = query.ExecuteNonQuery();
					}
				}
				using (MySqlCommand query = db.CreateCommand())
				{
					query.CommandType = CommandType.Text;
					query.CommandText = "update `feeds` set `last_refresh` = null where `feed_id` = @FeedId";
					query.Parameters.Clear();
					query.Parameters.AddWithValue("@FeedId", FeedId);
					query.Prepare();
					lock (lockDataReader)
					{
						affectedRows = query.ExecuteNonQuery();
					}
				}
				transaction.Commit();
			}
		}

		/// <summary>
		/// Get details about one requestedFeed
		/// </summary>
		/// <param name="Id"></param>
		/// <returns>The Feed</returns>
		public Feed GetFeed( long Id ) 
		{
			if (!db.Ping())
			{
				db.Open();
			}

			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText = 
					"select * " +
					"from `feeds` f " +
					"where f.`feed_id` = @FeedId";
				query.Parameters.Clear();
				query.Parameters.AddWithValue("@FeedId", Id);
				query.Prepare();

				lock (lockDataReader)
				{
					using (MySqlDataReader reader = query.ExecuteReader())
					{
						if (reader.HasRows)
						{
							if(reader.Read())
							{
								return new Feed(reader);
							}
						}
						reader.Close();
					}
				}
				return null;
			}
		}

		/// <summary>
		/// Create and/or update a requestedFeed
		/// </summary>
		/// <param name="requestedFeed"></param>
		/// <returns>Number of articles inserted</returns>
		public int SaveFeed(Feed feed)
		{
			// if( !FeedExists(feed.Url) )
			if( 0 == feed.FeedId )
			{
				feed.FeedId = CreateFeed(feed.Url, feed.Internet, feed.UpdateRate);
			}
			return UpdateFeed(feed);
		}

		/// <summary>
		/// Save a requestedFeed in the database
		/// </summary>
		/// <returns>Number of articles inserted</returns>
		public int UpdateFeed(Feed feed)
		{
			if (!db.Ping())
			{
				db.Open();
			}
			int affectedRows = 0;
			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText = 
					"update `feeds` set" +
					"`title` = @Title, "+
					"`url` = @Url, " +
					"`description` = @Description, "+
					"`picture` = @Picture, " +
					"`last_refresh` = now() " +
					"where `feed_id` = @FeedId";

				query.Parameters.AddWithValue( "@FeedId", feed.FeedId );
				query.Parameters.AddWithValue("@Url", feed.Url.AbsoluteUri);
				if( null == feed.Title ) 
				{
					query.Parameters.AddWithValue( "@Title", DBNull.Value );
				} else {
					query.Parameters.AddWithValue( "@Title", feed.Title );
				}
				if( null == feed.Description ) 
				{
					query.Parameters.AddWithValue( "@Description", DBNull.Value );
				} else {
					query.Parameters.AddWithValue( "@Description", feed.Description );
				}
				if( null == feed.Picture ) 
				{
					query.Parameters.AddWithValue( "@Picture", DBNull.Value );
				} else {
					query.Parameters.AddWithValue("@Picture", feed.Picture.AbsoluteUri);
				}
				/*if (null == feed.LastUpdated)
				{
					query.Parameters.AddWithValue("@LastUpdated", DateTime.Now);
				}
				else
				{
					query.Parameters.AddWithValue("@LastUpdated", feed.LastUpdated);
				}*/
				query.Prepare();
				lock (lockDataReader)
				{
					affectedRows = query.ExecuteNonQuery();
				}

				int affectedArticles = 0;
				if (null != feed.Articles)
				{
					foreach (Article article in feed.Articles)
					{
						if (null == article.Link)
						{
							continue;
						}
						if (!ArticleExists(article.Link, feed.FeedId))
						{
							SaveArticle(article);
							affectedArticles++;
						}
					}
				}
				return affectedArticles;
			}
		}

		/// <summary>
		/// Save an article
		/// </summary>
		/// <param name="article"></param>
		/// <returns>The Article Id</returns>
		public long SaveArticle(Article article)
		{
			if (!db.Ping())
			{
				db.Open();
			}
			int affectedRows = 0;
			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText = 
					"replace into `articles` (`ts_article`, `feed_id`, `date`, `title`, `headline`, `description`, `link`, `comments`, `author`) " +
					"values (now(), @FeedId, @Date, @Title, @Headline, @Description, @Link, @Comments, @Author)";
				query.Parameters.AddWithValue("@FeedId", article.FeedId);
				
				if (null == article.Date || DateTime.MinValue == article.Date )
				{
					query.Parameters.AddWithValue("@Date", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Date", article.Date);
				}
				
				if (null == article.Title || 0 == article.Title.Length)
				{
					query.Parameters.AddWithValue("@Title", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Title", article.Title);
				}
				
				if (null == article.Headline || 0 == article.Headline.Length)
				{
					query.Parameters.AddWithValue("@Headline", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Headline", article.Headline);
				}
				
				if (null == article.Description || 0 == article.Description.Length)
				{
					query.Parameters.AddWithValue("@Description", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Description", article.Description);
				}

				if (null == article.Link || 0 == article.Link.AbsoluteUri.Length)
				{
					query.Parameters.AddWithValue("@Link", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Link", article.Link);
				}
				
				if (null == article.Comments || 0 == article.Comments.Length)
				{
					query.Parameters.AddWithValue("@Comments", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Comments", article.Comments);
				}
				
				if (null == article.Author || 0 == article.Author.Length)
				{
					query.Parameters.AddWithValue("@Author", DBNull.Value);
				}
				else
				{
					query.Parameters.AddWithValue("@Author", article.Author);
				}
				query.Prepare();

				lock (lockDataReader)
				{
					affectedRows = query.ExecuteNonQuery();
					article.ArticleId = query.LastInsertedId;
				}
				return article.ArticleId;
			}
		}
		/// <summary>
		/// Get a specific article
		/// </summary>
		/// <param name="Id"></param>
		/// <returns>The article</returns>
		public Article GetArticle(long Id)
		{
			if (!db.Ping())
			{
				db.Open();
			}

			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText =
					"select * " +
					"from `articles` a " +
					"where a.`article_id` = @ArticleId";
				query.Parameters.Clear();
				query.Parameters.AddWithValue("@ArticleId", Id);
				query.Prepare();

				lock (lockDataReader)
				{
					using (MySqlDataReader reader = query.ExecuteReader())
					{
						if (reader.HasRows)
						{
							if (reader.Read())
							{
								return new Article(reader);
							}
						}
						reader.Close();
					}
				}
				return null;
			}
		}

		/// <summary>
		/// Check if a specific article exists in the database
		/// </summary>
		/// <param name="articleLink"></param>
		/// <param name="FeedId"></param>
		/// <returns>True if the article exists, false otherwise</returns>
		public bool ArticleExists(Uri articleLink, long FeedId)
		{
			if (!db.Ping())
			{
				db.Open();
			}

			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText =
					"select a.`article_id` " +
					"from `articles` a " +
					"where a.`link` = @Link " +
					"and a.`feed_id` = @FeedId";
				query.Parameters.Clear();
				query.Parameters.AddWithValue("@Link", articleLink.AbsoluteUri);
				query.Parameters.AddWithValue("@FeedId", FeedId);
				query.Prepare();

				lock (lockDataReader)
				{
					using (MySqlDataReader reader = query.ExecuteReader())
					{
						bool hasRows = reader.HasRows;
						reader.Close();
						return hasRows;
					}
				}
			}
		}

		/// <summary>
		/// Check if a specific requestedFeed exists in the database
		/// </summary>
		/// <param name="feedUrl"></param>
		/// <returns>True if the requestedFeed exists, false otherwise</returns>
		public bool FeedExists(Uri feedUrl)
		{
			if (!db.Ping())
			{
				db.Open();
			}

			using (MySqlCommand query = db.CreateCommand())
			{
				query.CommandType = CommandType.Text;
				query.CommandText =
					"select f.`feed_id` " +
					"from `feeds` f " +
					"where f.`url` = @Url";
				query.Parameters.Clear();
				query.Parameters.AddWithValue("@Url", feedUrl.AbsoluteUri);
				query.Prepare();

				lock (lockDataReader)
				{
					using (MySqlDataReader reader = query.ExecuteReader())
					{
						bool hasRows = reader.HasRows;
						reader.Close();
						return hasRows;
					}
				}
			}
		}
	}
}
