﻿/*
 * mAggregatorStatus - © EUROPEAN COMMISSION, 2009-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la
«Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que
conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse
suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le
logiciel distribué sous la Licence est distribué «en
l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT,
expresses ou implicites.
 * Consultez la Licence pour les autorisations et les
restrictions linguistiques spécifiques relevant de la
Licence.
 *
 * Auteur de l'oeuvre: Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Configuration;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.IO.Pipes;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Binary;

namespace mAggregatorService
{
	public partial class mAggregatorService : ServiceBase
	{
		/// <summary>
		/// Service's Main Thread
		/// </summary>
		Thread thMainThread;

		/// <summary>
		/// Status reporting thread
		/// </summary>
		Thread thStatusThread;

		/// <summary>
		/// Event triggered by the service's OnStop() method when the process should terminate itself
		/// </summary>
		private static ManualResetEvent stopService;

		/// <summary>
		/// Downloader Threads
		/// </summary>
		List<Thread> downloadThreads = new List<Thread>();
		/// <summary>
		/// Saver Threads
		/// </summary>
		List<Thread> saveThreads = new List<Thread>();

		/// <summary>
		/// Counters (for the status thread)
		/// </summary>
		mCounterSet Counters = new mCounterSet();

		/// <summary>
		/// Constructor
		/// </summary>
		public mAggregatorService()
		{
			InitializeComponent();
			CanStop = true;
			CanShutdown = true;
			CanPauseAndContinue = false;
		}

		/// <summary>
		/// Called on Service startup
		/// </summary>
		/// <param name="args"></param>
		protected override void OnStart(string[] args)
		{
#if DEBUG
			Debugger.Launch();
#endif
			LoadConfiguration();

			stopService = new ManualResetEvent(false);

			ParameterizedThreadStart startDlThread = new ParameterizedThreadStart(ThreadDownload);
			for (int id = 0; id < cfgNumDownloaders; id++)
			{
				downloadThreads.Add(new Thread(startDlThread));
				downloadThreads.Last().Start(new ThreadParamDownload());
			}

			ParameterizedThreadStart startSvThread = new ParameterizedThreadStart(ThreadSave);
			for (int iv = 0; iv < cfgNumSavers; iv++)
			{
				saveThreads.Add(new Thread(startSvThread));
				saveThreads.Last().Start(new ThreadParamSave());
			}

			ParameterizedThreadStart startStThread = new ParameterizedThreadStart(ThreadStatus);
			thStatusThread = new Thread(startStThread);
			thStatusThread.Start(new mThreadParam());

			ThreadStart fct = new ThreadStart(ThreadMain);
			thMainThread = new Thread(fct);
			thMainThread.Start();
		}

		/// <summary>
		/// Called on when service is requested to stop
		/// </summary>
		protected override void OnStop()
		{
			stopService.Set();

			foreach (Thread thread in downloadThreads.Union(saveThreads))
			{
				if (thread.IsAlive)
				{
					thread.Join(10000);
				}
			}
			if (thStatusThread.IsAlive)
			{
				thStatusThread.Join(1000);
			}
			if (thMainThread.IsAlive)
			{
				thMainThread.Join(5000);
			}
		}

		/// <summary>
		/// Main thread of the service
		/// </summary>
		private void ThreadMain()
		{
			bool commitSuicide = false;
			try
			{
				DateTime lastRefresh = new DateTime(0);
				while (false == stopService.WaitOne(cfgYieldTime))
				{
					using (FeedDatabase db = new FeedDatabase(cfgDbConnectionString, Encoding.GetEncoding(cfgDbTextEncoding)))
					{
						if (0 == lastRefresh.Ticks || cfgPollInterval < (DateTime.Now - lastRefresh).TotalMilliseconds)
						{
							Feed[] feeds = db.GetOutdatedFeeds();
							foreach (Feed feed in feeds)
							{
								Counters.Increment("DownloadRequest");
								EnqueueDownloadRequest(feed);
							}
							lastRefresh = DateTime.Now;
						}
					}
				}
			}
			catch (Exception e)
			{
				eventLog.WriteEntry("MainThread: An exception has occured " + e.ToString(), EventLogEntryType.Error);
				commitSuicide = true;
			}
			finally
			{
				stopService.Set();
			}

			if (commitSuicide)
			{
				Environment.Exit(1);
			}
			// Thread.CurrentThread.Abort();
		}

		/// <summary>
		/// Thread responsible of sending status notification to the listeners (if any)
		/// </summary>
		/// <param name="param"></param>
		private void ThreadStatus(object param)
		{
			mThreadParam arguments = (mThreadParam)param;
			String threadSignature = "StatusThread#" + arguments.InstanceNumber;
			Counters.Increment("ThreadStatus");
			if (true == cfgReportStatus)
			{
				DateTime lastRefresh = new DateTime(0);
				while (!stopService.WaitOne(cfgYieldTime))
				{
					try
					{
						if (0 == lastRefresh.Ticks || 1000 < (DateTime.Now - lastRefresh).TotalMilliseconds)
						{
							// Serialize the "Counters" object and send it to the listeners
							using (NamedPipeClientStream pipe = new NamedPipeClientStream(cfgStatusPipeServer, cfgStatusPipeName, PipeDirection.Out))
							{
								pipe.Connect(0);
								if (pipe.CanWrite)
								{
									BinaryFormatter bf = new BinaryFormatter();
									bf.Serialize(pipe, Counters.GetCountersCopy());
									pipe.Flush();
								}
							}
							lastRefresh = DateTime.Now;
						}
					}
					catch(System.TimeoutException) 
					{
						// Nothing to do (connection to pipe lost)
					}
					catch (Exception e)
					{
						eventLog.WriteEntry(threadSignature + ": " + e.ToString(), EventLogEntryType.Error);
					}
				}
			}
			Counters.Decrement("ThreadStatus");
		}

		/// <summary>
		/// Thread responsible for downloading documents
		/// </summary>
		/// <param name="param"></param>
		public void ThreadDownload(object param)
		{
			ThreadParamDownload arguments = (ThreadParamDownload)param;
			String threadSignature = "DownloadThread#" + arguments.InstanceNumber;

			Counters.Increment("ThreadDownload");
			try
			{
				while (!stopService.WaitOne(cfgYieldTime))
				{
					Feed feed = DequeueDownloadRequest();
					if (null != feed)
					{
						using (FeedDatabase db = new FeedDatabase(cfgDbConnectionString, Encoding.GetEncoding(cfgDbTextEncoding)))
						{
							// Placeholders for last exception data
							string leMessage = string.Empty;
							string leType = string.Empty;
							string leError = string.Empty;

							Counters.Increment("ActiveDownload");
							Counters.Decrement("DownloadRequest");
							try
							{
								threadSignature = "DownloadThread#" + arguments.InstanceNumber + "\r\nFeed#" + feed.FeedId + "\r\nUrl:" + feed.Url + "\r\n\r\n";
								// feed = db.GetFeed(arguments.Feed.FeedId);
								int nb = feed.Download(feed.Internet ? internetSettings : intranetSettings);
								Counters.Increment("SaveRequest");
								Counters.Increment("DownloadedDocuments");
								Counters.Increment("ParsedArticles", nb);

								EnqueueSaveRequest(feed);
							}
							catch (System.UriFormatException e)
							{
								leError = e.ToString();
								leType = e.GetType().Name;
								leMessage = e.Message;
							}
							catch (System.Net.WebException e)
							{
								leError = e.ToString();
								leType = e.GetType().Name;
								leMessage = e.Message;
							}
							catch (System.InvalidOperationException e)
							{
								leError = e.ToString();
								leType = e.GetType().Name;
								leMessage = e.Message;
							}
							catch (System.Xml.XmlException e)
							{
								leError = e.ToString();
								leType = e.GetType().Name;
								leMessage = e.Message;
							}
							finally
							{
								Counters.Decrement("ActiveDownload");
								if (leError != string.Empty)
								{
									Counters.Increment("DownloadErrors");
									string lePrefix = "[mAggregator::";
									string oldDescription = feed.Description;
									string newDescription = lePrefix + leType + leMessage;
									if (null != feed.Description && feed.Description.StartsWith(lePrefix))
									{
										if (oldDescription != newDescription)
										{
											eventLog.WriteEntry(threadSignature + ": " + leError, EventLogEntryType.Warning);
										}
									}
									else
									{
										eventLog.WriteEntry(threadSignature + ": " + leError, EventLogEntryType.Warning);
									}
									feed.Description = newDescription;
									db.UpdateFeed(feed);
								}
							}
						}
					}
				}
			}
			catch (Exception e)
			{
				string leError = threadSignature + ": " + e.ToString();
				eventLog.WriteEntry(threadSignature + ": " + leError, EventLogEntryType.Error );
			}
			finally
			{
				Counters.Decrement("ThreadDownload");
			}
		}

		/// <summary>
		/// Thread responsible for writing articles in the database
		/// </summary>
		/// <param name="param"></param>
		public void ThreadSave(object param)
		{
			ThreadParamSave arguments = (ThreadParamSave)param;
			String threadSignature = "SaveThread#" + arguments.InstanceNumber;

			Counters.Increment("ThreadSave");
			try
			{
				while (!stopService.WaitOne(cfgYieldTime))
				{
					Feed feed = DequeueSaveRequest();
					if (null != feed)
					{
						using (FeedDatabase db = new FeedDatabase(cfgDbConnectionString, Encoding.GetEncoding(cfgDbTextEncoding)))
						{
							Counters.Increment("ActiveSave");
							Counters.Decrement("SaveRequest");
							feed.Articles = feed.Articles.Where(f => f.Link != null).ToArray();
							try
							{
								threadSignature = "SaveThread#" + arguments.InstanceNumber + " Feed#" + feed.FeedId;
								if (feed.Articles.Length == 0)
								{
									db.TouchFeed(feed.FeedId);
								}
								else
								{
									int nb = db.SaveFeed(feed);
									Counters.Increment("SavedArticles", nb);
								}
							}
							catch (MySql.Data.MySqlClient.MySqlException e)
							{
								eventLog.WriteEntry(threadSignature + ": " + e.ToString(), EventLogEntryType.Warning);
								Counters.Increment("SaveErrors");
							}
							finally
							{
								Counters.Decrement("ActiveSave");
							}
						}
					}
				}
			}
			catch (Exception e)
			{
				eventLog.WriteEntry(threadSignature + ": " + e.ToString(), EventLogEntryType.Error);
			}
			finally
			{
				Counters.Decrement("ThreadSave");
			}
		}

		/// <summary>
		/// Type of object passed as argument to the Downloader Thread
		/// </summary>
		public class ThreadParamDownload : mThreadParam
		{
			private Feed _feed;

			public Feed Feed
			{
				get
				{
					return _feed;
				}
			}
			public ThreadParamDownload()
				: base()
			{
			}
			public ThreadParamDownload(Feed Feed)
				: base()
			{
				this._feed = Feed;
			}
		}

		/// <summary>
		/// Type of object passed as argument to the Saver Thread
		/// </summary>
		public class ThreadParamSave : mThreadParam
		{
			/// <summary>
			/// Feed to save into the database
			/// </summary>
			private Feed _feed;

			/// <summary>
			/// Feed to save into the database
			/// </summary>
			public Feed Feed
			{
				get
				{
					return _feed;
				}
			}
			public ThreadParamSave()
				: base()
			{
			}
			public ThreadParamSave(Feed Feed)
				: base()
			{
				this._feed = Feed;
			}
		}

		/// <summary>
		/// Object for synchronising access to the SaveQueue
		/// </summary>
		private object SaveQueueLocker = new object();
		/// <summary>
		/// Initializes the SaveQueue
		/// </summary>
		private volatile Queue<Feed> SaveQueue = new Queue<Feed>();
		/// <summary>
		/// Post a request to one of the SaverThread
		/// </summary>
		/// <param name="feed"></param>
		protected void EnqueueSaveRequest(Feed feed)
		{
			lock (SaveQueueLocker)
			{
				SaveQueue.Enqueue(feed);
			}
		}
		/// <summary>
		/// Fetch a request from the queue
		/// </summary>
		/// <returns></returns>
		protected Feed DequeueSaveRequest()
		{
			if (SaveQueue.Count > 0)
			{
				lock (SaveQueueLocker)
				{
					if (SaveQueue.Count > 0)
					{
						return SaveQueue.Dequeue();
					}
				}
			}
			return null;
		}

		/// <summary>
		/// Object for synchronising access to the DownloadQueue
		/// </summary>
		private object DownloadQueueLocker = new object();
		/// <summary>
		/// Initializes the DownloadQueue
		/// </summary>
		private volatile Queue<Feed> DownloadQueue = new Queue<Feed>();
		/// <summary>
		/// Post a request to one of the DownloaderThread
		/// </summary>
		/// <param name="feed"></param>
		protected void EnqueueDownloadRequest(Feed feed)
		{
			lock (DownloadQueueLocker)
			{
				DownloadQueue.Enqueue(feed);
			}
		}
		/// <summary>
		/// Fetch a request from the queue
		/// </summary>
		/// <returns></returns>
		protected Feed DequeueDownloadRequest()
		{
			if (DownloadQueue.Count > 0)
			{
				lock (DownloadQueueLocker)
				{
					if (DownloadQueue.Count > 0)
					{
						return DownloadQueue.Dequeue();
					}
				}
			}
			return null;
		}

	}
}
