﻿/*
 * © EUROPEAN COMMISSION, 2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous 
 * la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou 
 * implicites.
 * Consultez la Licence pour les autorisations et les restrictions 
 * linguistiques spécifiques relevant de la Licence.
 *
 * Auteur de l'oeuvre: Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace mFileMonitorService
{
	/// <summary>
	/// Extended FilesystemMonitor class containing data about the input and output folders
	/// </summary>
	public class FileMover : MetaNull.IO.FilesystemMonitor
	{
		/// <summary>
		/// An error has occured
		/// </summary>
		/// <param name="Error">Error description</param>
		public delegate void FilesystemErrorHandler(Exception Error);

		/// <summary>
		/// Triggered when an operation on the filesystem failed
		/// </summary>
		public event FilesystemErrorHandler OnFilesystemError;

		/// <summary>
		/// A file has been copied
		/// </summary>
		/// <param name="path">The file's path</param>
		public delegate void InputFileCopiedHandler(string path);

		/// <summary>
		/// Triggered when a file is being copied
		/// </summary>
		public event InputFileCopiedHandler OnInputFileCopied;

		/// <summary>
		/// A file has been deleted
		/// </summary>
		/// <param name="path">The file's path</param>
		public delegate void InputFileDeletedHandler(string path);

		/// <summary>
		/// Triggered when a file is being deleted
		/// </summary>
		public event InputFileDeletedHandler OnInputFileDeleted;


		/// <summary>
		/// An error has occured while processing the file
		/// </summary>
		/// <param name="Error">Error description</param>
		public delegate void ProcessingErrorHandler(Exception Error);

		/// <summary>
		/// Triggered when a processing on a file fails
		/// </summary>
		public event ProcessingErrorHandler OnProcessingError;


		/// <summary>
		/// A file is being processed by an external process
		/// </summary>
		/// <param name="path">The file's path</param>
		/// <param name="processPath">The process' path</param>
		/// <param name="processArguments">The process' arguments</param>
		public delegate void InputFileProcessingHandler(string path, string processPath, string processArguments);

		/// <summary>
		/// Triggered when a file is being processed by an external process
		/// </summary>
		public event InputFileProcessingHandler OnInputFileProcessing;

		/// <summary>
		/// A file has been processed by an external process
		/// </summary>
		/// <param name="path">The file's path</param>
		/// <param name="processPath">The process' path</param>
		/// <param name="processArguments">The process' arguments</param>
		/// <param name="exitCode">The exit code of the process</param>
		/// <param name="processOutput">The output of the process (STDOUT)</param>
		/// <param name="processError">The output of the process (STDERROR)</param>
		public delegate void InputFileProcessedHandler(string path, string processPath, string processArguments, int exitCode, string processOutput, string processError);

		/// <summary>
		/// Triggered when a file has been processed
		/// </summary>
		public event InputFileProcessedHandler OnInputFileProcessed;


		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="input">The input folder</param>
		/// <param name="outputs">List of outputs folders for the input files</param>
		public FileMover(Input input, Output[] outputs)
			: base(input.Path, input.Filter)
		{
			this._input = input;
			this._outputs = outputs;
			this.IncludeSubdirectories = input.IncludeSubFolders;

			OnFileCreated += new FileCreatedHandler(OnFileCreatedHandler);
			OnFileRenamed += new FileRenamedHandler(OnFileRenamedHandler);
		}

		private Input _input;
		/// <summary>
		/// The input folder
		/// </summary>
		public Input input
		{
			get
			{
				return _input;
			}
		}

		private Output[] _outputs;
		/// <summary>
		/// The input folders (no need to filter them as the handler function 
		/// will ensure that only output folders matching the specified input will be used)
		/// </summary>
		public Output[] outputs
		{
			get
			{
				return _outputs;
			}
		}

		/// <summary>
		/// Handle renamed files events
		/// </summary>
		/// <param name="OldFullPath">Old file name</param>
		/// <param name="NewFullPath">New file name</param>
		void OnFileRenamedHandler(string OldFullPath, string NewFullPath)
		{
			// Copy/Move?
			OnFileCreatedHandler(NewFullPath);
		}

		/// <summary>
		/// Replace specific placeholder strings into a filepath string
		/// </summary>
		/// <param name="subject">The string to search for placeholders</param>
		/// <param name="RelativePath">relative path of the input file (used for some replacements)</param>
		/// <param name="FullPath">full path of the input file (used for some replacements)</param>
		/// <returns></returns>
		private static string PathReplacePlaceholders( string subject, string RelativePath, string FullPath ) {
			DateTime inputDT = System.IO.File.GetCreationTime(FullPath);
			return subject
				.Replace("%input-path%", FullPath)
				.Replace("%input-relative-path%", RelativePath)
				.Replace("%input-directory%", System.IO.Path.GetDirectoryName(FullPath))
				.Replace("%input-file%", System.IO.Path.GetFileName(FullPath))
				.Replace("%input-file-no-ext%", System.IO.Path.GetFileNameWithoutExtension(FullPath))
				.Replace("%input-extension%", System.IO.Path.GetExtension(FullPath))
                .Replace("%input-date%", String.Format("{0:0000}",inputDT.Year)+String.Format("{0:00}",inputDT.Month)+String.Format("{0:00}",inputDT.Day))
				.Replace("%date%", String.Format("{0:0000}", DateTime.Now.Year) + String.Format("{0:00}", DateTime.Now.Month) + String.Format("{0:00}", DateTime.Now.Day))
				.Replace("%input-time%", String.Format("{0:00}", inputDT.Hour) + String.Format("{0:00}", inputDT.Minute) + String.Format("{0:00}", inputDT.Second))
				.Replace("%time%", String.Format("{0:00}", DateTime.Now.Hour) + String.Format("{0:00}", DateTime.Now.Minute) + String.Format("{0:00}", DateTime.Now.Second))
				;
		}
		
		/// <summary>
		/// Handles created files events
		/// </summary>
		/// <param name="FullPath">Full path of the file</param>
		void OnFileCreatedHandler(string FullPath)
		{
			string fileBase = System.IO.Path.GetDirectoryName(FullPath);
			// Double check that the file belongs to the Input we are handeling
			if (!fileBase.StartsWith(input.Path))
			{
				if (null != OnFilesystemError)
				{
					OnFilesystemError(new ArgumentException("File " + FullPath + " does not belong to " + input.Path + " base: " + fileBase));
				}
			}
			else
			{
				string relativePath = System.IO.Path.GetFileName(FullPath);
				if (input.Path != fileBase)
				{
					relativePath = System.IO.Path.Combine(fileBase.Substring(input.Path.Length + 1), System.IO.Path.GetFileName(FullPath));
				}
				// Browse the available outputs for the current input file
				foreach (Output output in outputs.Where(output => output.InputNames.Contains(input.Name) || output.InputNames.Contains("*")).ToArray())
				{

					if (output.Path.StartsWith("process://") || output.Path.StartsWith("process-output-to-file://") || output.Path.StartsWith("process-error-to-file://"))
					{
						try {
							string processPath = string.Empty;
							string processArguments = string.Empty;
							string processTargetPath = string.Empty;
							int offset = output.Path.IndexOf("://") + 3;
							if (output.Path.StartsWith("process://"))
							{
								string[] processData = output.Path.Substring(offset).Split(new char[] { ' ' }, 2);
								processPath = processData[0];
								if (processData.Length > 1)
								{
									processArguments = FileMover.PathReplacePlaceholders(processData[1], relativePath, FullPath);
								}
							}
							else
							{
								
								string[] processData = output.Path.Substring(offset).Split(new char[] { ' ' }, 3);
								if (processData.Length < 2)
								{
									throw new ArgumentException("process-output-to-file:// and process-error-to-file:// expects at elast two arguments" );
								}
								processTargetPath = processArguments = FileMover.PathReplacePlaceholders(processData[0], relativePath, FullPath);
								if (null != OnInputFileProcessed)
								{
									OnInputFileProcessing(processTargetPath, string.Empty, string.Empty);
								}
								processPath = processData[1];
								if (processData.Length > 2)
								{
									processArguments = FileMover.PathReplacePlaceholders(processData[2], relativePath, FullPath);
								}
							}
							if (null != OnInputFileProcessing)
							{
								OnInputFileProcessing(input.Path, processPath, processArguments );
							}
							System.Diagnostics.Process proc = new System.Diagnostics.Process();
							proc.StartInfo.FileName = processPath;
							proc.StartInfo.Arguments = processArguments;
							proc.StartInfo.WindowStyle = System.Diagnostics.ProcessWindowStyle.Hidden;
							proc.StartInfo.UseShellExecute = false;
							proc.StartInfo.RedirectStandardError = true;
							proc.StartInfo.RedirectStandardOutput = true;
							proc.Start();
							string stdOutput = proc.StandardOutput.ReadToEnd();
							string stdError = proc.StandardError.ReadToEnd();
							proc.WaitForExit();
							int exitCode = proc.ExitCode;
							proc.Close();

							if (output.Path.StartsWith("process-output-to-file://") || output.Path.StartsWith("process-output-to-file://"))
							{
								System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(processTargetPath));
								using (System.IO.FileStream fs = System.IO.File.Create(processTargetPath))
								{
									using( System.IO.StreamWriter sw = new System.IO.StreamWriter(fs) ) {
										if (output.Path.StartsWith("process-output-to-file://"))
										{
											sw.Write(stdOutput);
										}
										else
										{
											sw.Write(stdError);
										}
										sw.Close();
									}
								}
							}

							if (null != OnInputFileProcessed)
							{
								OnInputFileProcessed(input.Path, processPath, processArguments, exitCode, stdOutput, stdError);
							}
						}
						catch (Exception e)
						{
							if (null != OnProcessingError)
							{
								OnProcessingError(e);
							}
						}
					}
					else
					{
						try
						{
							string target = string.Empty;
							// If CopyTree is set, duplicate the realtive path into the output directory
							// Otherwise just put the file in the root
							if (output.CopyTree)
							{
								target = System.IO.Path.Combine(output.Path, relativePath);
							}
							else
							{
								target = System.IO.Path.Combine(output.Path, System.IO.Path.GetFileName(FullPath));
							}
							// Make sure that output directory exists
							System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(target));
							// Copy the file (will be deleted further if required)
							if (null != OnInputFileCopied)
							{
								OnInputFileCopied(FullPath);
							}
							System.IO.File.Copy(FullPath, target, true);
						}
						catch (Exception e)
						{
							if (null != OnFilesystemError)
							{
								OnFilesystemError(e);
							}
						}
					}
					
				}
				// Remove the original file if requested
				if (!input.KeepOriginal)
				{
					try
					{
						if (System.IO.File.Exists(FullPath))
						{
							System.IO.File.Delete(FullPath);
							if (null != OnInputFileDeleted)
							{
								OnInputFileDeleted(FullPath);
							}
						}
					}
					catch (Exception e)
					{
						if (null != OnFilesystemError)
						{
							OnFilesystemError(e);
						}
					}
				}
			}
		}
	}
}
