<?php
/*
 * © EUROPEAN COMMISSION, 2007-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou implicites.
 * Consultez la Licence pour les autorisations et les restrictions linguistiques spécifiques relevant de la Licence.
 *
 * Auteurs:
 *	Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *	Cosmin Popescu,   Adrian.POPESCU@ext.ec.europa.eu
 */

namespace MetaNull\Application {

	/**
	 * Base class defining the mechanism of an application
	 * @author Pascal Havelange
	 * @copyright © EUROPEAN COMMISSION, 2008
	 * @package MetaNull\Application
	 * @license http://ec.europa.eu/idabc/eupl
	 */
	abstract class mBaseApplication {

		/**
		 * Application name
		 * @var string
		 */
		protected $app_name = '';

		/**
		 * Application version
		 * @var string
		 */
		protected $app_version = '1.0';

		/**
		 * Application Authors
		 * @var array
		 */
		protected $app_authors = array();

		/**
		 * Application Copyright
		 * @var string
		 */
		protected $app_copyright = '';

		/**
		 * Object Constructor
		 */
		public function __construct() {
		}

		/**
		 * This method must be implemented in the derivated class
		 * It is called first (prior to the onExecute() method)
		 * Throw an Exception from this method to immediatelly stop the application (in which case onDestroy() will be gracefuly called)
		 * @access protected
		 */
		protected function onInitialize() {}

		/**
		 * This method must be implemented in the derivated class
		 * This method is called last (after the onExecute() method)
		 * No exception should be thrown within this method, its only goal is to possibly free some resources
		 * @access protected
		 */
		protected function onDestroy() {}

		/**
		 * This method must be implemented in the derivated class
		 * Main logic of the application. Throws an application from within this method to exit the application
		 * @access protected
		 * @abstract
		 */
		abstract protected function onExecute();

		/**
		 * This method must be implemented in the derivated class
		 * This method is called whenever an exceptions occurs in onInitialize or onExecute
		 * Please note that the onDestroy() method will still be called afterwards.
		 * @access protected
		 */
		protected function onException( $e ) {
			trigger_error( (string)$e );
		}

		/**
		 * Core method of the application.
		 * @access public
		 * @final
		 */
		final public function run() {
			try {
				$this->onInitialize();
				$this->onExecute();
			}
			catch (\MetaNull\Tools\mysqlException $e){
				$this->onException( $e );
			}
			catch( \Exception $e ) {
				$this->onException( $e );
			}
			$this->onDestroy();
		}

		/**
		 * Returns the name of the application
		 * @access public
		 * @return string
		 * @final
		 */
		final public function getName() {
			return $this->app_name;
		}

		/**
		 * Returns the version of the application
		 * @access public
		 * @return string
		 */
		final public function getVersion() {
			return $this->app_version;
		}

		/**
		 * Returns the names of the authors of the application
		 * @access public
		 * @return array An array of author names
		 * @final
		 */
		final public function getAuthors() {
			return $this->app_authors();
		}

		/**
		 * Returns the copyright notice of the application
		 * @access public
		 * @return string
		 * @final
		 */
		final public function getCopyright() {
			return $this->app_copyright;
		}
	}

}
?>