<?php
/*
 * © EUROPEAN COMMISSION, 2007-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou implicites.
 * Consultez la Licence pour les autorisations et les restrictions linguistiques spécifiques relevant de la Licence.
 *
 * Auteurs:
 *	Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *	Cosmin Popescu,   Adrian.POPESCU@ext.ec.europa.eu
 */

namespace MetaNull\Commission {

	/**
	 * Classes establishing an HTTP Session, and authenticating the user with the ECAS system
	 * iAuthentication implementation focussed on the usage of the ECAS system
	 * @author Pascal Havelange
 	 * @copyright © EUROPEAN COMMISSION, 2008
 	 * @package MetaNull\Commission
	 * @license http://ec.europa.eu/idabc/eupl
	 */
	class EC_Authentication implements \MetaNull\Authentication\iAuthentication {
		
		/**
		 * Base URL of the application (Used by the login system during redirections)
		 * @var string
		 */
		protected $application_url = false;
		
		/**
		 * Base URL of the application, with the last query string (Used by the login system during redirections)
		 * @var string
		 */
		protected $application_url_with_query = false;
		
		/**
		 * Name of the HTTP_SESSION
		 * @var string
		 */
		protected $session_name = false;
		
		/**
		 * Full path to the SSL Certificate
		 * @var string
		 */
		protected $ssl_certificate = false;
		
		/**
		 * Constructor
		 * 
		 * @param string $session_name HTTP Session Name
		 * @param string $ssl_certificate Full Path to the SSL Certificate
		 * @param string $application_url Base URL of your application (where ECAS has to redirect the user in case of successful authentication)
		 * @access public
		 */
		public function __construct( $session_name, $application_url, $ssl_certificate ) {
                    if( !file_exists( $ssl_certificate )) {
                            throw new \MetaNull\Authentication\authenticationException( "Invalid SSL Certificate path: {$ssl_certificate}" );
                    }
                    $this->ssl_certificate = $ssl_certificate;

                    if(  false == $application_url || 0 == strlen( $application_url )) {
                            throw new \MetaNull\Authentication\authenticationException( "Invalid argument: \$application_url" );
                    }
                    $this->application_url = rawurlencode( $application_url );
                    $this->application_url_with_query = $this->application_url;
                    if( isset( $_SERVER['QUERY_STRING']) && 0 != strlen($_SERVER['QUERY_STRING'])
                            && false === preg_match( '/[&^]ticket=/i', $_SERVER['QUERY_STRING'] )	// DO NOT SEND THE TICKET TWICE!
                    ) {
                            $query_string = preg_replace(
                                    '/(^ticket=[a-zA-Z0-9=_\-]+$|^ticket=[a-zA-Z0-9=_\-]+&|&ticket=[a-zA-Z0-9=_\-]+$|&ticket=[a-zA-Z0-9=_\-]+&)/i', '&', $_SERVER['QUERY_STRING'] );
                            if( 0 != strlen($query_string) && '&' != $query_string ) {
                                    $this->application_url_with_query .= rawurlencode("?{$query_string}");
                            }
                    }
                    if( !is_string( $session_name ) || 0 == strlen( $session_name )) {
                            throw new \MetaNull\Authentication\authenticationException( "Invalid argument: \$session_name" );
                    }
                    $this->session_name = preg_replace( '/[^[:alnum:]]/i', '_', $session_name );
		}
		
		/**
		 * Closes a session and - if requested - logoff the user
		 * @param bool $logoff If set to True, also logoff the user
		 */
		public function endSession( $logoff = false ) {
			$_SESSION = array();
			if( isset( $_COOKIE[session_name()])) {
				setcookie( session_name(), '', time() - 42000, '/' );
			}
			session_destroy();
			if( $logoff ) {
				$this->call_ecas_logoff_page();
			}
		}
		
		/**
		 * Check if the session has been established successfully
		 * @return bool True if the section has been established, false otherwise
		 */
		public function sessionEstablished() {
			return isset( $_SESSION['ecasuserid'] ) ? true : false;
		}
		
		
		/**
		 * If needed, redirect to the ECAS login page
		 * @access public
		 */
		final protected function call_ecas_login_page() {
			if( headers_sent() ) {
				throw new ecasException( 'HTTP Headers already sent!' );
			}
			$url = $this->application_url;
			/*if (array_key_exists('QUERY_STRING', $_SERVER)){
				$url .= urlencode('?' . $_SERVER['QUERY_STRING']);
			}*/
			header( 'Location: ' . self::EC_ECAS_SERVER . "/login?service={$url}&TariqaPRO=" . session_id());
			exit();
		}
		
		/**
		 * Ends up the HTTP Session and Log off from ECAS
		 * @access public
		 */
		final protected function call_ecas_logoff_page() {
			if( headers_sent() ) {
				throw new ecasException( 'HTTP Headers already sent!' );
			}
			header( 'Location: ' . self::EC_ECAS_SERVER . "/logout?service={$this->application_url}" );
			exit();
		}
		
		/**
		 * Check the validity of an ECAS "ticket"
		 * 
		 * @param string $ticket ECAS Ticket ID
		 * @return string|bool The User ID in case of success, false otherwise
		 * @final
		 * @access protected
		 */
		final protected function checkECASTicket( $ticket ) {
			$check_url = self::EC_ECAS_SERVER . "/validate?ticket={$ticket}&service={$this->application_url}";
			$http_error = false;
			$http_response = \MetaNull\Tools\mCurl::Get( $check_url, $http_error, $this->ssl_certificate );
			if( false === $http_response || 200 != $http_error ) {
				throw new ecasException( "HTTP Error #{$http_error}:\n{$http_response}", $http_error );
			}
			$response_lines = explode("\n", $http_response );
			$ecas_authenticated = $response_lines[0];
			$ecas_userid = $response_lines[1];
			$this->last_authentication_strength = isset( $response_lines[2] ) ? $response_lines[2] : false;
			
			return ('yes' != strtolower( $ecas_authenticated )) ? false : $ecas_userid;
		}

	}
	
}
?>