<?php
/*
 * © EUROPEAN COMMISSION, 2007-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou implicites.
 * Consultez la Licence pour les autorisations et les restrictions linguistiques spécifiques relevant de la Licence.
 *
 * Auteurs:
 *	Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *	Cosmin Popescu,   Adrian.POPESCU@ext.ec.europa.eu
 */

namespace MetaNull\News{
	
	/**
	 * Class for transforming the query string from unified language to engine specific language.
	 * @author Cosmin Popescu
	 * @copyright © EUROPEAN COMMISSION, 2009-2010
	 * @package MetaNull\News
	 * @license http://ec.europa.eu/idabc/eupl
	 */
	class unifiedQuerySyntax {

		/**
		 * The unified query syntax type
		 * @var string
		 
		 */
		const UNIFIED_QUERY_SYNTAX = 'UQS';

		/**
		 * The original syntax type
		 * @var string
		 
		 */
		const ORIGINAL_SYNTAX = 'OS';

		/**
		 * The news configuration for translating the query
		 * @var array
		 * @access private
		 */
		private $newsConfig;

		/**
		 * The query string
		 * @var string
		 * @access private
		 */
		private $query;

		/**
		 * The key for the translation
		 * @var string
		 * @access private
		 */
		private $translation_key;

		/**
		 * An array containing the operators
		 * @var string[]
		 * @access private
		 */
		private $operators;

		/**
		 * An array containing the functions
		 * @var string[]
		 * @access private
		 */
		private $functions;

		/**
		 * The delimiter (for keywords)
		 * @var string
		 * @access private
		 */
		private $delimiter = '"';

		/**
		 * The left paranthesis
		 * @var string
		 * @access private
		 */
		private $leftP = "(";

		/**
		 * The right paranthesis
		 * @var string
		 * @access private
		 */
		private $rightP = ")";

		/**
		 * The literals keywords (between quotes)
		 * @var string[]
		 * @access private
		 */
		private $literal_keywords = array();

		/**
		 * The separators (a list of characters that we are sure that will not be found in a query string)
		 * @var string[]
		 * @access private
		 */
		private $separators = array("\x01", "\x02", "\x307");

		/**
		 * The list of reserved characters (characters used by regular expression which should be replaced during the translation and then replaced back
		 * @var string[]
		 * @access private
		 */
		private $resChars = array("-" => "MINUS");

		/**
		 * The implicit operator (if a keyword does not have an operator before, this will be added
		 * @var string
		 * @access private
		 */
		private $implicitOperator = "or";

		/**
		 * The dates conditions (These could only appear at the end)
		 * @var string
		 * @access private
		 */
		private $date_conditions = null;

		/**
		 * Function that will be called by preg_replace to replace the literal string from the query string
		 * @param string[] $matches The matches
		 * @access private
		 * @return string
		 */
		private function getLiterals($matches) {
			$i = count($this->literal_keywords);
			$this->literal_keywords[$i] = $matches[0];

			$sep = $this->separators[1];

			return $sep . $i . $sep;
		}

		/**
		 * Gets the functions from the news config file
		 * @access protected
		 * @return string[]
		 */
		protected function getFunctions() {
			$functions = array();
			if (array_key_exists("functions", $this->newsConfig)) {
				foreach ($this->newsConfig["functions"] as $key => $function) {
					$functions[] = str_replace("-", " ", $key);
				}
			}

			return $functions;
		}

		/**
		 * Gets the operators from the news config file
		 * @access protected
		 * @return string[]
		 */
		protected function getOperators() {
			$operators = array();
			if (array_key_exists("operators", $this->newsConfig)) {
				foreach ($this->newsConfig["operators"] as $key => $operator) {
					$operators[] = str_replace("-", " ", $key);
				}
			}

			return $operators;
		}

		/**
		 * Gets the delimiter from the news config file
		 * @access protected
		 * @return string
		 */
		protected function getDelimiter() {
			$delimiter = null;
			if (array_key_exists("delimiter", $this->newsConfig)) {
				if (array_key_exists("@", $this->newsConfig["delimiter"])) {
					if (array_key_exists("value", $this->newsConfig["delimiter"]["@"])) {
						$delimiter = $this->newsConfig["delimiter"]["@"]["value"];
					}
				}
			}

			return $delimiter;
		}

		/**
		 * Gets the parantheses from the news config file
		 * @access protected
		 * @return string[]
		 */
		protected function getParantheses() {
			$parantheses = array(null, null);
			if (array_key_exists("parantheses", $this->newsConfig)) {
				$p = $this->newsConfig["parantheses"];
				if (array_key_exists("left", $p)) {
					if (array_key_exists("@", $p["left"])) {
						if (array_key_exists("value", $p["left"]["@"])) {
							$parantheses[0] = $p["left"]["@"]["value"];
						}
					}
				}
				if (array_key_exists("right", $p)) {
					if (array_key_exists("@", $p["right"])) {
						if (array_key_exists("value", $p["right"]["@"])) {
							$parantheses[1] = $p["right"]["@"]["value"];
						}
					}
				}
			}

			return $parantheses;
		}

		/**
		 * Converts a date from any string format to a UNIX timestamp
		 * @param string $date The date
		 * @access public
		 * @static
		 * @return int
		 */
		public static function timstamp($date) {
			if (preg_match('/[\d]+([ \-\.])[\d]+([ \-\.])[\d]+.*$/', $date, $matches)){
				$date = str_replace($matches[1], '/', $date);
				$date = str_replace($matches[2], '/', $date);
			}
			$d = \DateTime::createFromFormat('d/m/Y', $date);
			if ($d === false){
				$d = \DateTime::createFromFormat('d/m/y', $date);
				if ($d === false){
					$d = \DateTime::createFromFormat('m/d/Y', $date);
					if ($d === false){
						$d = \DateTime::createFromFormat('m/d/y', $date);
						if ($d === false){
							$d = \DateTime::createFromFormat('Y/m/d', $date);
							if ($d === false){
								$this->throwError($date);
								throw new unifiedQueryException("Unified Query Syntax error: Date not well formated. ");
							}
						}
					}
				}
			}
			/* @var $d DateTime */
			return $d->getTimestamp();
		}

		/**
		 * Extracts the dates conditions from the query
		 * Returns an associative array with the query and the date condition separated
		 * @param string $query The query string
		 * @access public
		 * @static
		 * @return array
		 */
		public static function getDatesCondition($query) {
			$pattern1 = '/((range)\(([\d\-\/\.]+)[ \n\t\r]*\,[ \n\t\r]*([\d\-\/\.]+)\))[ \n\t\r\)]*$/';
			$pattern2 = '/((previous)\(([\d]+)\))[ \n\t\r\)]*$/';

			if (preg_match($pattern1, $query, $matches) || preg_match($pattern2, $query, $matches)){
				return array(
					'query' => str_replace($matches[1], '', $query),
					'string' => $matches[1],
					'function' => $matches[2],
					'params' => ($matches[2] == 'range'?array($matches[3], $matches[4]):array($matches[3])), 
				);
			}

			return array('query' => $query);

		}

		/**
		 * Constructs the object
		 * @param string $query The query string
		 * @param array $config The news configuration for translating the query
		 * @param string $translation_key The key for the translation
		 * @access public
		 * @return void
		 */
		public function __construct($query, $config, $translation_key) {
			if (array_key_exists('search', $config))
				$this->newsConfig = $config['search'];
			else $this->newsConfig = $config;
			$this->translation_key = $translation_key;
			foreach ($this->resChars as $key => $ch) {
				$this->resChars[$key] = $this->separators[2] . $ch . $this->separators[2];
			}
			$this->query = $query;
			$dc = self::getDatesCondition($this->query);
			if (array_key_exists('string', $dc)){
				$this->query = $dc['query'];
				$this->date_conditions = $dc['string'];
			}
			$this->query = $this->escapeArray($this->query);
			$this->operators = $this->escapeArray($this->getOperators());
			$this->functions = $this->escapeArray($this->getFunctions());
			$this->delimiter = $this->escapeArray($this->getDelimiter());
			$parantheses = $this->getParantheses();
			$this->leftP = $this->escapeArray($parantheses[0]);
			$this->rightP = $this->escapeArray($parantheses[1]);
		}

		/**
		 * Escapes the reserved characters for regular expressions patterns (like - | etc.). See the resChars array
		 * @param string $s The string to escape
		 * @access private
		 * @return string
		 */
		private function escape($s) {
			foreach ($this->resChars as $key => $value) {
				$s = str_replace($key, $value, $s);
			}

			return $s;
		}

		/**
		 * Escapes the reserved characters for an array of values or a string
		 * @param mixed $values The values
		 * @access private
		 * @return mixed
		 */
		private function escapeArray($values) {
			if (is_array($values)) {
				foreach ($values as $key => $value) {
					$values[$key] = $this->escape($value);
				}

				return $values;
			}
			else {
				return $this->escape($values);
			}
		}

		/**
		 * Unescapes a reserved characters for regular expressions (like - | etc.). See the resChars array
		 * @param string $s The string to unescape
		 * @access private
		 * @return string
		 */
		private function unescape($s) {
			foreach ($this->resChars as $key => $value) {
				$s = str_replace($value . " ", $key, $s);
				$s = str_replace($value, $key, $s);
			}

			return $s;
		}

		/**
		 * Unescapes the reserved characters for an array of values or a string
		 * @param mixed $values The values
		 * @access private
		 * @return mixed
		 */
		private function unescapeArray($values) {
			if (is_array($values)) {
				foreach ($values as $key => $value) {
					$values[$key] = $this->unescape($value);
				}

				return $values;
			}
			else {
				return $this->unescape($values);
			}
		}

		/**
		 * Returns the tokens for parsing the query string
		 * @param string $query The query string [OTPIONAL]
		 * @param string[] $operators An array containing the operators [OTPIONAL]
		 * @param string[] $functions An array containing the functions [OTPIONAL]
		 * @param string $delimiter The string delimiters [OPTIONAL]
		 * @param string $leftP The left paranthesis character [OPTIONAL]
		 * @param string $rightP The right paranthesis character [OPTIONAL]
		 * @access private
		 * @return queryToken[]
		 */
		private function queryTokens($query = null, $operators = null, $functions = null, $delimiter = null, $leftP = null, $rightP = null) {
			//The algorithm used is Reversed Polish Expressions

			//Set the query, the operators, the functions, the delimiter, the parantheses and the separator

			if ($query == null) {
				$query = $this->query;
			}
			
			if ($operators == null) {
				$operators = $this->operators;
			}
			if ($functions == null) $functions = $this->functions;
			if ($delimiter == null) $delimiter = $this->delimiter;
			if ($leftP == null) $leftP = $this->leftP;
			if ($rightP == null) $rightP = $this->rightP;

			$sep = $this->separators[0];
			$sep2 = $this->separators[1];

			//First we will split with regular expressions in tokens.
			//In order not to parse something between quotes (the texts between quotes should be search as they are),
			//we take out everything that is in the quotes, put it into an array

			if ($delimiter != null) {
				$pattern = '/(' . $delimiter . '[^' . $delimiter . ']*' . $delimiter . ')/';
				$query = preg_replace_callback($pattern, array($this, "getLiterals"), $query);
			}

			//Search for spaces in the operators
			//If we have spaces in the operators we should take them out (it is possibile to have "and" and "and not".
			//If we don't take out the spaces, then the and not keyword will be parsed as operator "and" and keyword "not keyword",
			//instead of operator "and not" and keyword "keyword"
			$space_operators = array();
			foreach ($operators as $key => $operator) {
				if (strpos($operator, " ") !== false) {
					$op = str_replace(" ", "", $operator);
					//We should also keep track of replaced spaces, so in the end we would be able to put the spaces back
					//After all, the operator is "and not" and not "andnot"
					$space_operators[$op] = $operator;
					//Replace in the query
					$query = str_replace($operator, $op, $query);
					//Replace in the operators list
					$operators[$key] = $op;
				}
			}

			//Setting the patterns and the replacements.
			$patterns = array();
			$repl = array();
			foreach ($functions as $function) {
				$patterns[] = '/\b(' . $function . ')\b/';
				$repl[] = $sep . queryToken::TOKEN_TYPE_FUNC . '\1' . $sep;
			}
			foreach ($operators as $operator) {
				$patterns[] = '/\b(' . $operator . ')\b/';
				$repl[] = $sep . queryToken::TOKEN_TYPE_OPERATOR . '\1' . $sep;
			}

			$patterns[] = '(' . "\\" . $leftP . ')';
			$repl[] = $sep . queryToken::TOKEN_TYPE_LPAR . $leftP . $sep;
			$patterns[] = '(' . "\\" . $rightP . ')';
			$repl[] = $sep . queryToken::TOKEN_TYPE_RPAR . $rightP . $sep;

			//Try to identify into the query the keywords, operators, functions etc. and put them between certain characters
			//we are sure that we will not find \x01 in a query string
			$query = preg_replace($patterns, $repl, $query);

			//If we have a quote set up, then put back the strings between the quotes.
			if ($delimiter != null) {
				foreach ($this->literal_keywords as $key => $keyword) {
					$query = str_replace($sep2 . $key . $sep2, $keyword, $query);
				}
			}

			//Now we have the string with every operator, function, parantheses between the separator (\x01).
			//What we are doing next, is to split the string after the delimiter. Bassically, every odd index should be an operator, paranthese or function
			//and what is on an even position should be a keyword.

			//So create the tokens...
			$parts = explode($sep, $query);
			$parts = array_map( 'trim', $parts );
			$parts = array_filter( $parts );

			$tokens = array();

			$last_token = null;

			foreach ($parts as $i => $tok) {
				if ($i % 2 == 1) {
					$prec = null;
					$type = substr($tok, 0, queryToken::TYPE_LENGTH);
					$value = substr($tok, queryToken::TYPE_LENGTH);
					if ($last_token != null) {
						if (($last_token->type == queryToken::TOKEN_TYPE_KEYWORD && $type != queryToken::TOKEN_TYPE_OPERATOR && $type != queryToken::TOKEN_TYPE_RPAR) ||
								($last_token->type == queryToken::TOKEN_TYPE_RPAR && ($type == queryToken::TOKEN_TYPE_KEYWORD || $type == queryToken::TOKEN_TYPE_FUNC))) {
							$key = null;
							foreach($operators as $key => $v) break;
							if ($key !== null) {
								$token = new queryToken(queryToken::TOKEN_TYPE_OPERATOR, $operators[$key]);
								$tokens[] = $token;
							}
						}
					}
					if ($type == queryToken::TOKEN_TYPE_OPERATOR) {
						if (array_key_exists($value, $space_operators)) $value = $space_operators[$value];
						foreach($operators as $key => $op) {
							if ($op == $value) {
								$prec = $key;
								break;
							}
						}
					}
					$tokens[] = new queryToken($type, $value, $prec);
				}
				else {
					if ($last_token != null) {
						if ($last_token->type == queryToken::TOKEN_TYPE_KEYWORD || $last_token->type == queryToken::TOKEN_TYPE_RPAR) {
							$key = null;
							foreach($operators as $key => $v) break;
							if ($key !== null) {
								$token = new queryToken(queryToken::TOKEN_TYPE_OPERATOR, $operators[$key]);
								$tokens[] = $token;
							}
						}
					}

					if ($delimiter != null && substr($tok, 0, strlen($delimiter)) != $delimiter && substr($tok, -strlen($delimiter)) != $delimiter) {
						$words = explode(" ", $tok);
						$last_word = null;
						foreach ($words as $word) {
							if ($word != "") {
								if ($last_word != null) {
									$key = null;
									foreach($operators as $key => $v) break;
									if ($key !== null) {
										$token = new queryToken(queryToken::TOKEN_TYPE_OPERATOR, $operators[$key]);
										$tokens[] = $token;
									}
								}
								$tokens[] = new queryToken(queryToken::TOKEN_TYPE_KEYWORD, $word);
								$last_word = $word;
							}
						}
					}
					else {
						$tokens[] = new queryToken(queryToken::TOKEN_TYPE_KEYWORD, $tok);
					}
				}
				$keys = array_keys($tokens);
				$key = end($keys);
				$last_token = $tokens[$key];
			}

			return $tokens;
		}

		/**
		 * Throw a token error
		 * @param queryToken $token The token with the problem
		 * @access private
		 * @return void
		 */
		private function throwError($token) {
			ob_start();

			var_dump($token);

			$contents = ob_get_contents();
			ob_end_clean();
			error_log($contents);
		}

		/**
		 * Function performing Shunting Yard algotithm
		 * @param queryToken[] $tokens A string array containing the tokens
		 * @access private
		 * @return queryToken[]
		 */
		private function shuntingYard($tokens) {
			//This is an implementation of the shunting - yard algorithm.
			//The tokens are supposed to be in the polish form
			$ops = array();
			$output = array();

			foreach ($tokens as $token) {
				if ($token->type == queryToken::TOKEN_TYPE_KEYWORD) {
					$output[] = $token;
				}
				else if ($token->type == queryToken::TOKEN_TYPE_FUNC) {
					$ops[] = $token;
				}
				else if ($token->type == queryToken::TOKEN_TYPE_SEPARATOR) {
					if (count($ops) == 0) {
						$this->throwError($token);
						throw new unifiedQueryException("Unified Query Syntax error: A separator was misplaced. See the log for details. ");
					}
					$_t = array_pop($ops);
					while ($_t->type != queryToken::TOKEN_TYPE_LPAR) {
						if (count($ops) == 0) {
							$this->throwError($token);
							throw new unifiedQueryException("Unified Query Syntax error: A separator was misplaced or some parentheses were mismatched. See the log for details. ");
						}
						$output[] = $_t;
						$_t = array_pop($ops);
					}
					$ops[] = $_t;
				}
				else if ($token->type == queryToken::TOKEN_TYPE_OPERATOR) {
					if (count($ops) > 0) {
						$_t = array_pop($ops);
						while ($_t->type == queryToken::TOKEN_TYPE_OPERATOR && $token->precedence <= $_t->precedence) {
							$output[] = $_t;
							$_t = null;
							if (count($ops) == 0) {
								break;
							}
							$_t = array_pop($ops);
						}
						if ($_t != null) {
							$ops[] = $_t;
						}
					}
					$ops[] = $token;
				}
				else if ($token->type == queryToken::TOKEN_TYPE_LPAR) {
					$ops[] = $token;
				}
				else if ($token->type == queryToken::TOKEN_TYPE_RPAR) {
					if (count($ops) == 0) {
						$this->throwError($token);
						throw new unifiedQueryException("Unified Query Syntax error: Missplaced right paranthesis. See the log for details. ");
					}
					$_t = array_pop($ops);
					while ($_t->type != queryToken::TOKEN_TYPE_LPAR) {
						$output[] = $_t;
						if (count($ops) == 0) {
							$this->throwError($token);
							throw new unifiedQueryException("Unified Query Syntax error: Missplaced right paranthesis. See the log for details. ");
						}
						$_t = array_pop($ops);
					}

					if (count($ops) > 0) {
						$_t = array_pop($ops);
						if ($_t->type == queryToken::TOKEN_TYPE_FUNC) {
							$output[] = $_t;
						}
						else {
							$ops[] = $_t;
						}
					}
				}
			}

			while (count($ops) > 0) {
				$_t = array_pop($ops);
				if ($_t->type == queryToken::TOKEN_TYPE_LPAR) {
					$this->throwError($_t);
					throw new unifiedQueryException("Unified Query Syntax error: Missplaced paranthesis. See the log for details. ");
				}
				else if ($_t->type == queryToken::TOKEN_TYPE_OPERATOR) {
					$output[] = $_t;
				}
				else {
					$this->throwError($_t);
					throw new unifiedQueryException("Unified Query Syntax error: Missplaced tokens. See the log for more details. ");
				}
			}

			return $output;
		}

		/**
		 * Splits the string in tokens
		 * @param string $query The query string
		 * @access protected
		 * @return queryToken[]
		 */
		protected function tokenize($query) {
			return $this->shuntingYard($this->queryTokens($query));
		}

		/**
		 * Normalizes the tokens if there is a default function (no keyword can be independent) - it addes this function after each independent keywork
		 * @param queryToken $tokens The tokens
		 * @param string $function The function to be added (the name)
		 * @access private
		 * @return queryToken[]
		 */
		private function normalize($tokens, $function) {
			$stack = array();
			foreach ($tokens as $key => $token) {
				if ($token->type == queryToken::TOKEN_TYPE_KEYWORD) {
					$stack[] = array($key);
				}
				else if ($token->type == queryToken::TOKEN_TYPE_OPERATOR) {
					if (count($stack) < 2) return $tokens;
					$key1 = array_pop($stack);
					$key2 = array_pop($stack);

					if (is_array($key1) || is_array($key2)) {
						$stack[] = array_merge((is_array($key1)?$key1:array()), (is_array($key2)?$key2:array()));
					}
					else {
						$stack[] = $token->value;
					}
				}
				else if ($token->type == queryToken::TOKEN_TYPE_FUNC) {
					if (count($stack) < 1) return $tokens;
					$keys = array_pop($stack);
					$stack[] = $token->value;
				}
				else return $tokens;
			}

			if (count($stack) != 1) return $tokens;
			$keys = array_pop($stack);
			if (!is_array($keys)) return $tokens;

			$result = array();
			foreach ($tokens as $key => $token) {
				$result[] = $token;
				if (array_search($key, $keys) !== false) {
					$result[] = new queryToken(queryToken::TOKEN_TYPE_FUNC, $function);
				}
			}

			return $result;
		}

		/**
		 * Returns true if a query left paranthesis contains engine reserved word like 'like'
		 * @param string $leftP The paranthesis characters
		 * @access private
		 * @return boolean
		 */
		private function like($leftP) {
			return substr($leftP, 0, 5) == " like";
		}

		/**
		 * Gets the translated values for the functions
		 * @param string[] $originals The original functions values
		 * @access protected
		 * @return string[]
		 */
		protected function getTranslatedFunctions($originals) {
			$functions = array();
			foreach ($originals as $function) {
				$functions[$function] = "";
				$f = str_replace(" ", "-", $function);
				if (array_key_exists("translate", $this->newsConfig["functions"][$f])) {
					if (array_key_exists($this->translation_key, $this->newsConfig["functions"][$f]["translate"])) {
						if (array_key_exists("@", $this->newsConfig["functions"][$f]["translate"][$this->translation_key])) {
							if (array_key_exists("value", $this->newsConfig["functions"][$f]["translate"][$this->translation_key]["@"])) {
								$functions[$function] = $this->newsConfig["functions"][$f]["translate"][$this->translation_key]["@"]["value"];
							}
							if (array_key_exists("default", $this->newsConfig["functions"][$f]["translate"][$this->translation_key]["@"])) {
								$functions["default"] = $function;
							}
						}
					}
				}
			}

			return $functions;
		}

		/**
		 * Gets the translated values for the operators
		 * @param string[] $originals The original operators values
		 * @access protected
		 * @return string[]
		 */
		protected function getTranslatedOperators($originals) {
			$operators = array();
			foreach ($originals as $operator) {
				$operators[$operator] = "";
				$o = str_replace(" ", "-", $operator);
				if (array_key_exists("translate", $this->newsConfig["operators"][$o])) {
					if (array_key_exists($this->translation_key, $this->newsConfig["operators"][$o]["translate"])) {
						if (array_key_exists("@", $this->newsConfig["operators"][$o]["translate"][$this->translation_key])) {
							if (array_key_exists("value", $this->newsConfig["operators"][$o]["translate"][$this->translation_key]["@"])) {
								$operators[$operator] = $this->newsConfig["operators"][$o]["translate"][$this->translation_key]["@"]["value"];
							}
						}
					}
				}
			}

			return $operators;
		}

		/**
		 * Gets the translated values for the delimiter
		 * @param string $original The original delimiter
		 * @access protected
		 * @return string
		 */
		protected function getTranslatedDelimiter($original) {
			$delimiter = null;

			if (array_key_exists("delimiter", $this->newsConfig)) {
				if (array_key_exists("translate", $this->newsConfig["delimiter"])) {
					if (array_key_exists($this->translation_key, $this->newsConfig["delimiter"]["translate"])) {
						if (array_key_exists("@", $this->newsConfig["delimiter"]["translate"][$this->translation_key])) {
							if (array_key_exists("value", $this->newsConfig["delimiter"]["translate"][$this->translation_key]["@"])) {
								$delimiter = $this->newsConfig["delimiter"]["translate"][$this->translation_key]["@"]["value"];
							}
						}
					}
				}
			}
			return $delimiter;
		}

		/**
		 * Gets the translated values for the parantheses
		 * @param string[] $originals The original parantheses
		 * @access protected
		 * @return string[]
		 */
		protected function getTranslatedParantheses($originals) {
			$parantheses = array("", "");

			if (array_key_exists("parantheses", $this->newsConfig)) {
				$p = $this->newsConfig["parantheses"];
				if (array_key_exists("left", $p)) {
					if (array_key_exists("translate", $p["left"])) {
						if (array_key_exists($this->translation_key, $p["left"]["translate"])) {
							if (array_key_exists("@", $p["left"]["translate"][$this->translation_key])) {
								if (array_key_exists("value", $p["left"]["translate"][$this->translation_key]["@"])) {
									$parantheses[0] = $p["left"]["translate"][$this->translation_key]["@"]["value"];
								}
							}
						}
					}
				}

				if (array_key_exists("right", $p)) {
					if (array_key_exists("translate", $p["right"])) {
						if (array_key_exists($this->translation_key, $p["right"]["translate"])) {
							if (array_key_exists("@", $p["right"]["translate"][$this->translation_key])) {
								if (array_key_exists("value", $p["right"]["translate"][$this->translation_key]["@"])) {
									$parantheses[1] = $p["right"]["translate"][$this->translation_key]["@"]["value"];
								}
							}
						}
					}
				}
			}

			return $parantheses;
		}

		/**
		 * Translates a query from Unified Query Language to Specific Engine Language
		 * @param queryToken[] $tokens The tokens to be translated [OPTIONAL]
		 * @access public
		 * @return string
		 */
		public function translate($tokens = null) {
			if (preg_match('/^[ \t\n\r]*$/', $this->query) && $this->date_conditions != null){
				return $this->date_conditions;
			}
			//Here, the tokens are allready in the polish form, so we translate the string
			//First we escape any character from the operators, functions, delimiter, parantheses - in order to be able to parse them using regular expression
			$operators = $this->escapeArray($this->getTranslatedOperators($this->getOperators()));
			$functions = $this->escapeArray($this->getTranslatedFunctions($this->getFunctions()));
			$delimiter = $this->escapeArray($this->getTranslatedDelimiter($this->getDelimiter()));
			$paranthesis = $this->getTranslatedParantheses($this->getParantheses());
			$leftP = $this->escapeArray($paranthesis[0]);
			$rightP = $this->escapeArray($paranthesis[1]);

			if ($tokens == null) $tokens = $this->tokenize($this->query);

			//Normalizes the tokens if there is a default function (no keyword can be independent, like for the LatestNews engine)
			//We add this function after each independent keywork

			if (array_key_exists("default", $functions)) {
				if ($functions["default"] != null) $tokens = $this->normalize($tokens, $functions["default"]);
			}

			$stack = array();
			$output = "";

			foreach ($tokens as $token) {
				$type = $token->type;
				//If the token is a keyword, just put it on the stack
				if ($type == queryToken::TOKEN_TYPE_KEYWORD || $type == queryToken::TOKEN_TYPE_RESULT_KEYWORD) {
					$stack[] = $token;
				}
				//If the token is a functions,
				else if ($type == queryToken::TOKEN_TYPE_FUNC) {
					if (count($stack) == 0) {
						$this->throwError($token);
						throw new unifiedQueryException("Unified Query Syntax error: Missplaced function");
					}
					//Extract the number of function parameters from the stack (usually 1, but who knows?)
					$_t = array_pop($stack);
					//If the left parantheses contains like (see LatestNews)
					if ($this->like($leftP)) {
						//If there is the case, remove the delimiter
						//This is for the case title("hillary clinton candidature") put in the LatestNews engine.
						//If we don't take out the quote (which in this case is the delimiter),
						//The result would be like title like '%"hillary clinton candidature"%' which is obviously wrong.
						//The result should be title like '%hillary clinton candidature%'
						$v = $_t->value;
						if (substr($v, 0, strlen($delimiter)) == $delimiter) {
							$v = substr($v, 1);
						}
						if (substr($v, -strlen($delimiter)) == $delimiter) {
							$v = substr($v, 0, strlen($v) - 1);
						}

						//And, of course, escape the value
						$value = mysql_escape_string($v);
					}
					else {
						$value = $_t->value;
					}
					//If we have a right paranthese
					if (($rightP != "" && $rightP != null && !$this->like($leftP)) || $_t->type != queryToken::TOKEN_TYPE_RESULT_KEYWORD) {
						$_delimiter = (($functions[$token->value] == "" || $functions[$token->value] == null)?'"':'');
						//If the function translation is not null, translate using that functions
						if ($functions[$token->value] != "" && $functions[$token->value] != null) {
							$result = $functions[$token->value] . $leftP . $_delimiter . $value . $_delimiter . $rightP;
						}
						else {
							//Otherwise, just put the value
							//This is for the cases like for example LexisNexis, which do not have the function source.
							//In this case, if the query is title(obama) and source(reuters), we should translate reuters simply by reuters,
							//because we do not have a translation for the "source" function
							$result = $value;
						}
					}
					else {
						//If we don't have a right paranthese, then it means that we have to parse the value and to add the function in front of
						//each parameter
						//This is for the cases like google, where title(barack or obama) should be translated by
						//intitle:barack or intitle:obama, and not by intitle:barack or obama
						//So tokenize for the given expression (the current value of the token) with the same parameters but the functions, which is only our functions
						$__tokens = $this->queryTokens($_t->value, $operators, array($token->value), $this->delimiter, $this->leftP, $this->rightP);
						//After, apply the shunting yard algorithm to obtain the tokens in the polish form
						//For the above mentioned example, we would obtain the array {barack, obama, or}
						//We know for sure that we only have operators and keywords. This means we cannot have a query like title(barack and title(obama)). This we generate an error
						//So, knowing that, we cand after each keyword to put our function in order to have the function before every argument
						$__tokens = $this->shuntingYard($__tokens);
						$__out = array();
						foreach ($__tokens as $__token) {
							if ($__token->type == queryToken::TOKEN_TYPE_KEYWORD) {
								$__out[] = $__token;
								$__out[] = $token;
							}
							else if ($__token->type == queryToken::TOKEN_TYPE_OPERATOR) {
								$__out[] = $__token;
							}
							else {
								$this->throwError($token);
								throw new unifiedQueryException("Unified Query Syntax error: Unexpected function or paranthesis. ");
							}
						}

						//So, in the final, the tokens array will be (barack, title, obama, title, or)
						//Now, all we have to do is translate this expression, using the translated operators.
						//We cannot translate it as title:barack or title:obama. We have to translate it as title:barack OR title:obama
						//In general, search engines can be case sensitive when it comes to the operators. So in the final phase,
						//we should also have the translated operator

						$__operators = array();
						foreach ($operators as $op) {
							$__operators[$op] = $op;
						}
						$value = $this->translate($__operators, $functions, $delimiter, $leftP, $rightP, $__out);
						$_delimiter = "";
						$result = $value;
					}
					$_t = new queryToken(queryToken::TOKEN_TYPE_RESULT_KEYWORD, $result);
					$stack[] = $_t;
				}
				//If the token is an operator
				else if ($type == queryToken::TOKEN_TYPE_OPERATOR) {
					if (count($stack) < 2) {
						$this->throwError($token);
						throw new unifiedQueryException("Unified Query Syntax error: Missplaced operator");
					}
					$_t1 = array_pop($stack);
					$_t2 = array_pop($stack);

					//Pop the last 2 tokens from the stack, put between them the operator and push the result to the stack

					$result = ($_t2->type == queryToken::TOKEN_TYPE_RESULT_KEYWORD?"(":"") . $_t2->value . ($_t2->type == queryToken::TOKEN_TYPE_RESULT_KEYWORD?")":"") . " " .
							$operators[$token->value] . " " . ($_t1->type == queryToken::TOKEN_TYPE_RESULT_KEYWORD?"(":"") . $_t1->value . ($_t1->type == queryToken::TOKEN_TYPE_RESULT_KEYWORD?")":"");
					$_t = new queryToken(queryToken::TOKEN_TYPE_RESULT_KEYWORD, $result);
					$stack[] = $_t;
				}
			}

			//If we have more or less than one element in the stack in the final, then we have a problem.
			if (count($stack) != 1) {
				$this->throwError(null);
				throw new unifiedQueryException("Unified Query Syntax error: Invalid expression");
			}

			$token = array_pop($stack);

			$result = $this->unescape($token->value);

			if (substr($this->translation_key, 0, 6) == "Google") {
				$result = str_replace(array("(", ")"), "", $result);
			}

			if ($this->date_conditions != null){
				$result .= $this->date_conditions;
			}

			return $result;
		}
	}
}
?>