<?php
/*
 * © EUROPEAN COMMISSION, 2007-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou implicites.
 * Consultez la Licence pour les autorisations et les restrictions linguistiques spécifiques relevant de la Licence.
 *
 * Auteurs:
 *	Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *	Cosmin Popescu,   Adrian.POPESCU@ext.ec.europa.eu
 */

namespace MetaNull\Tools {

	if( !defined( 'APPLICATION_CACHE' )) {
		define( 'APPLICATION_CACHE', 'cache' );
	}
	if( !defined( 'APPLICATION_CACHE_VALIDITY' )) {
		define( 'APPLICATION_CACHE_VALIDITY', 3600 );	// 1 hour
	}
	if( !defined( 'APPLICATION_CACHE_USE' )) {
		define( 'APPLICATION_CACHE_USE', true );
	}

	/**
	 * Helper class enabling caching of random variables
	 * @author Cosmin Popescu, Pascal Havelange
	 * @copyright © EUROPEAN COMMISSION, 2010
	 * @package MetaNull\Tools
	 * @license http://ec.europa.eu/idabc/eupl
	 */
	class mCache {

		/**
		 * The application cache path
		 * @var string
		 * @access private
		*/
		private static $cachePath = null;
		
		/**
		 * Fetches the filename for a variable
		 
		 */
		public static function getVariableFileName( $name ) {
			return mCache::$cachePath . $name;
		}
		
		/**
		 * Initializes the application cache path
		 * @access private
	 * @static
		 * @return void
		*/
		private static function init_cache_path(){
			$fullpath = '';
			if( false != APPLICATION_CACHE ) {
				$fullpath .= preg_replace( "/[\\/\\\\]$/", '', APPLICATION_CACHE );
				$fullpath = preg_replace( "/[\\/\\\\]/", '/', $fullpath ) . '/';
			}
			self::$cachePath = $fullpath;
		}
		
		/**
		 * Caches a variable value
		 * @param mixed $value The variable value
		 * @param string $name The name of the cache data
		 * @param boolean $base64 If true, base64 encode the value after the serialization [OPTIONAL]
		 * @access public
	 * @static
		 * @return boolean
		*/
		public static function saveVariable($value, $name, $base64 = true){
			if( false !== APPLICATION_CACHE_USE ) {
				$fullpath = self::getVariableFileName($name);
				//Cache the variable
				$fSuccess = false;
				if( false !== ($f = fopen($fullpath, "wt"))) {
					$fSuccess = fwrite($f, ($base64 ? base64_encode(serialize($value)) : serialize($value)));
					fclose($f);
				}
				return $fSuccess;
			} else {
				return true;
			}
		}
		
		/**
		 * Deletes a cached variable
		 * @param string $name The variable name
		 * @access public
	 * @static
		 * @return void
		*/
		public static function deleteVariable($name){
			$fullpath = self::getVariableFileName($name);
			if (file_exists($fullpath)){
				@unlink($fullpath);
			}
		}
		
		/**
		 * Returns true if a variable is cached
		 * @param string $name The name of the variable
		 * @param string $file If this is set, then the cache time is compared with the file time and if the file is newer, the function will return false [OPTIONAL]
		 * This means that the variable is practically the content of a parsed file. 
		 * @access public
	 * @static
		 * @return boolean
		*/
		public static function isCached($name, $file = null){
			if( false === APPLICATION_CACHE_USE ) {
				return false;
			}
			$fullpath = self::getVariableFileName($name);
			
			if ($file !== null && @filemtime($file) > @filemtime($fullpath)) {
				return false;
			}
			
			$validity_condition = ((APPLICATION_CACHE_VALIDITY == -1)
				? false
				: (time() - APPLICATION_CACHE_VALIDITY) > @filemtime( $fullpath ));
			
			if( !file_exists( $fullpath ) || 
				$validity_condition ||
				0 == filesize( $fullpath )
			) {
				return false;
			} else {
				// Cache is valid, use it "as is"
				return true;
			}
		}
		
		/**
		 * Loads a variable from cache
		 * @param string $name The name of the cache data
		 * @param string $file If this is set, then the cache time is compared with the file time and if the file is newer, the function will return false [OPTIONAL]
		 * @param boolean $base64 If true, base64 decode the unserialized value [OPTIONAL]
		 * @access public
	 * @static
		 * @return mixed
		*/
		public static function loadVariable($name, $file = null, $base64 = true){
			if( false !== APPLICATION_CACHE_USE && self::isCached($name, $file)){
				//if ($name == 'countries/US') error_log(__LINE__.' Get the filename...');
				$fullpath = self::getVariableFileName($name);
				//if ($name == 'countries/US') error_log(__LINE__.' Get the content of the ' . $name . ' variable...');
				$content = file_get_contents($fullpath);
				//if ($name == 'countries/US') error_log(__LINE__.' Base 64 decode the ' . $name . ' variable...');
				$base64 = base64_decode($content);
				//if ($name == 'countries/US') error_log(__LINE__.' Unserialize the ' . $name . ' variable...');
				$result = unserialize($base64);
				//if ($name == 'countries/US') error_log(__LINE__.' Done with ' . $name . ' variable...');
				return $result;

				//return unserialize(($base64?base64_decode(file_get_contents($fullpath)):file_get_contents($fullpath)));
			}
			return null;
		}
		
		/**
		 * Adds a '/' at the end of a path if it is the case
		 * @param string $path The path
		 * @access public
	 * @static
		 * @return string
		*/
		public static function path($path){
			if (substr($path, -1) != '/' && substr($path, -1) != "\\") {
				return $path . '/';
			}
			return $path;
		}
		
		/**
		 * Initializes
		 * @param string $config_path The full path to the configuration file
		 * @param string $config_format The file format used by the configuration file
		 * @access public
	 * @static
		 * @return void
		*/
		public static function initialize( $config_path = null, $config_format = 'xml' ){
			self::init_cache_path();
		}
	}
	
	// Initializes the static properties
	mCache::initialize();
}

?>