<?php
/*
 * © EUROPEAN COMMISSION, 2007-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou implicites.
 * Consultez la Licence pour les autorisations et les restrictions linguistiques spécifiques relevant de la Licence.
 *
 * Auteurs:
 *	Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *	Cosmin Popescu,   Adrian.POPESCU@ext.ec.europa.eu
 */

namespace tariqa3 {

	/**
	 * Tarîqa 3.0 Application
	 * @package tariqa3
	 * @author Pascal Havelange, Cosmin Popescu
	 * @copyright © EUROPEAN COMMISSION, 2009-2010
	 * @license http://ec.europa.eu/idabc/eupl
	 */
	class Application extends \MetaNull\Application\mModularApplication {
		
		const MODULES_BASE_FOLDER = 'modules';

		const TRANSLATION_ROOT_SECTION = 'root';
		const TRANSLATION_MODULE_SECTION = 'module';

		/**
		 * Connection to the main Tarîqa 3.0 database
		 * @var \MetaNull\Tools\mMysql
		 */
		public $db;

		/**
		 * The whole translationTable for current language
		 * @var array
		 */
		protected $translationTable;

		/**
		 * The current language id
		 * @var string
		 */
		protected $currentLanguage;

		
		/**
		 * Constructs the application module
		 * @param array $application_config The application params. The array should be an associative array
		 *				If you want to send the config path and the config format just send an array like array('path' => 'my_path', 'format' => 'my_format')
		 *				If you want to send the config array directly, just send an array like array('config' => $my_config)
		 * @return void
		 */
		public function __construct( $application_config ) {
			try {
				parent::__construct( $application_config );
				$this->base_folder = '';

				// Connect to the database
				$this->db = @\tariqa3\statics::open_db_connection($this->config['database']);

				// Load translationTable
				if( array_key_exists( $this->config['application']['language-variable-name'], $_GET )) {
					$this->setCurrentLanguage($_GET[$this->config['application']['language-variable-name']]);
					setcookie($this->config['application']['language-variable-name'], $_GET[$this->config['application']['language-variable-name']], time() + 60 * 60 * 24 * 30);
				}
				else if( array_key_exists( $this->config['application']['language-variable-name'], $_COOKIE )) {
					$this->setCurrentLanguage($_COOKIE[$this->config['application']['language-variable-name']]);
				}
				else if( array_key_exists( $this->config['application']['language-variable-name'], $_REQUEST )) {
					$this->setCurrentLanguage($_REQUEST[$this->config['application']['language-variable-name']]);
				}
				else {
					$this->setCurrentLanguage($this->getDefaultLanguage());
				}
			} catch( \Exception $e ) {
				$this->onException($e);
			}
		}

		/**
		 * Checks PHP settings
		 * @return bool
		 */
		public static function checkPhpSettings() {
			if( 'On' == ini_get( 'register_globals' )) {
				trigger_error( 'Invalid php setting: register_globals should be Off' );
				return false;
			}

			if( !extension_loaded( 'mysqli' )) {
				trigger_error( 'Missing php module: mysqli' );
				return false;
			}

			return true;
		}

		/**
		 * Handle exceptions thrown by the modules
		 * @access protected
		 */
		protected function onException($e) {
			error_log((string)$e);

			$template = new \MetaNull\Application\MTemplate( 'm' );
			$template->loadTemplate( 'themes/fatalError.html' );

			$error_report =
				"\n" . '------------------------------' . "\n" .
				'Error report:'."\n".
				(string)$e .
				"\n" . '------------------------------' . "\n";

			$template->data['appName'] = $this->config['application']['name'];
			$template->data['recipient'] = $this->config['application']['contact'] .
					'?subject=' . rawurlencode('Bug Report') .
					'&body=' . rawurlencode($error_report);
			$template->data['error'] = '<strong>'.$e->getMessage() . '</strong>'."\n\n" . (string)$e;
			$template->data['content-type'] = 'Content-type: text/html; charset=utf-8';

			echo $template->transform();
			exit();
		}


		/**
		 * Returns the current translation table
		 * @param string $section Section Id
		 * @return array
		 */
		final public function getTranslationTable( $section = false ) {
			if( false === $section ) {
				return $this->translationTable;
			}
			if( array_key_exists($section, $this->translationTable)) {
				return $this->translationTable[$section];
			}
			return false;
		}

		/**
		 * Returns the current translation table
		 * @param string $module Key of the module
		 * @return array
		 */
		final public function getModuleTranslationTable( $module ) {
/*
 * DISABLE CACHING DURING DEVELOPMENT
 *
 *			$tmodule = 'translationTable_'.$module;
 *			$ttable = \MetaNull\Tools\mCache::loadVariable($tmodule);
 */
			$ttable = null;
/* * */
			if( null !== $ttable) {
				return $ttable;
			}

			$root_key = self::TRANSLATION_ROOT_SECTION;
			if( !array_key_exists($root_key, $this->translationTable )) {
				throw new \MetaNull\Application\applicationException('The root translation table (key: '.$root_key.') is empty');
			}
			$module_key = self::TRANSLATION_MODULE_SECTION . '/' . $module;
			if( !array_key_exists($module_key, $this->translationTable )) {
				throw new \MetaNull\Application\applicationException('The module translation table (key: '.$module_key.') is empty');
			}
			$table = $this->translationTable[$root_key];
			foreach( $this->translationTable[$module_key] as $key => $val ) {
				$new_key = self::TRANSLATION_MODULE_SECTION . '/' . $key;
				$table[$new_key] = $val;
			}

			\MetaNull\Tools\mCache::saveVariable($table, 'translationTable_' . $module);

			return $table;
		}

		/**
		 * Get the list of language id set in the configuration file (used for text translation)
		 * @return string[]
		 */
		final public function getLanguageList() {
			if( !array_key_exists('languages', $this->config['application']) &&
				!array_key_exists('language', $this->config['application']['languages'])
			) {
				throw new \InvalidArgumentException('No language list set in the configuration file!');
			}
			$list = array();
			foreach( $this->config['application']['languages']['language'] as $k => $v ) {
				if( is_array($v) ) {
					$list[] = $v['#'];
				} else {
					$list[] = $v;
				}
			}
			return $list;
		}
		/**
		 * Get the id of the default language as set in the configuration file
		 * @return string
		 */
		final public function getDefaultLanguage() {
			if( !array_key_exists('languages', $this->config['application']) &&
				!array_key_exists('language', $this->config['application']['languages'])
			) {
				throw new \InvalidArgumentException('No language list set in the configuration file!');
			}
			$list = array();
			foreach( $this->config['application']['languages']['language'] as $k => $v ) {
				if( is_array($v) &&
					array_key_exists( '@', $v ) &&
					array_key_exists( 'default', $v['@']) &&
					0 == strcasecmp( 'true', $v['@']['default'] )
				) {
					return $v['#'];
				}
			}
			throw new \InvalidArgumentException('No default language set in the configuration file!');
		}
		
		/**
		 * Get the id of the currently selected language
		 * @return string
		 */
		final public function getCurrentLanguage() {
			return $this->currentLanguage;
		}
		/**
		 * Set the id of the currently selected language (and reload the translation table)
		 * @param string $language Language id
		 */
		final public function setCurrentLanguage( $language ) {
			if( $this->currentLanguage != $language ) {
				$this->loadTranslationTable( $language );
			}
		}
		/**
		 * Load the translation table for a given language
		 * @param string $language Language Id
		 * @param string $section Section Id
		 * @return void;
		 */
		final public function loadTranslationTable( $language ) {
			$this->currentLanguage = $language;
/*
 * DISABLE CACHING DURING DEVELOPMENT
 *
 *			$ttable = \MetaNull\Tools\mCache::loadVariable('translationTable');
 */
			$ttable = null;
/* * */
			if( null !== $ttable) {
				$this->translationTable = $ttable;
				return;
			}

			$language_id = $this->db->real_escape_string( $language );
			$query = <<<sql
select s.`section_id`
from `languages_sections` s
sql;

			$local_table = array();
			$rc = $this->db->query( $query );
			if( false === $rc ) {
				throw new \MetaNull\Tools\mysqlException($this->db->error, $this->db->errno);
			}
			while( $row = $rc->fetch_assoc()) {
				$local_table[$row['section_id']] = array();
			}
			$rc->close();

			$query = <<<sql
select s.`section_id`, t.`keyword`, v.`data`
from `languages_sections` s, `translation` t, `translation_values` v
where s.`section_id` = t.`section_id`
and t.`section_id` = v.`section_id`
and t.`keyword` = v.`keyword`
and v.`language_id` = '{$language_id}'
and s.`section_id` = ?
sql;
			$stmt = $this->db->prepare( $query );
			if( false === $stmt ) {
				throw new \MetaNull\Tools\mysqlException($this->db->error, $this->db->errno);
			}
			foreach( $local_table as $section_id => $v ) {
				$stmt->bind_param( 's', $section_id );
				$stmt->execute();
				$stmt->bind_result($_section_id,$_keyword,$_data);
				$local_table[$section_id] = array();
				while( $stmt->fetch() ) {
					$local_table[$section_id][$_keyword] = $_data;
				}
				$stmt->free_result();
			}
			$stmt->close();
			$this->translationTable = $local_table;
		}

		/**
		 * Get the translated text corresponding to a key or to a set of keys
		 * @param string $section The section's id
		 * @param string|array|object $keys The key, an array of keys, or an object (properties used as keys)
		 * @return string|array|object Returns the tanslated text or an array or object containing translated text
		 */
		final public function translate( $section, $keys ) {
			if( is_string( $keys )) {
				return $this->translateInternals( $section, $keys );
			}
			if( is_array( $keys )) {
				$trslArray = array();
				foreach( $keys as $k => $v ) {
					$trslArray[$k] = $this->translate($section, $v);
				}
				return $trslArray;
			}
			if( is_object( $keys )) {
				$o = $this->translate($section, (array)$keys);
				foreach( $o as $k => $v ) {
					if( property_exists( get_class($keys), $k)) {
						$keys->$k = $v;
					}
				}
				return $keys;
			}
			return null;
		}
		/**
		 * Get the translated text associated to a specific key
		 * @param string $section The section's id
		 * @param string $key The text's key
		 * @return string
		 */
		final private function translateInternals( $section, $key ) {
			if( !is_string( $section )) {
				throw new \InvalidArgumentException( '$section argument must be of string type' );
			}
			if( !is_string( $key )) {
				throw new \InvalidArgumentException( '$keys argument must be of string type' );
			}
			if(!array_key_exists($section, $this->translationTable)) {
				return null;
			}
			if(!array_key_exists($key, $this->translationTable[$section])) {
				return null;
			}
			return $this->translationTable[$section][$key];
		}
	}
}

?>