<?php
/*
 * © EUROPEAN COMMISSION, 2007-2010
 *
 * Concédée sous licence EUPL, version 1.1 uniquement (la «Licence»).
 * Vous ne pouvez utiliser la présente oeuvre que conformément à la Licence.
 * Vous pouvez obtenir une copie de la Licence à l’adresse suivante:
 *
 * http://ec.europa.eu/idabc/eupl
 *
 * Sauf obligation légale ou contractuelle écrite, le logiciel distribué sous la Licence est distribué «en l’état»,
 * SANS GARANTIES OU CONDITIONS QUELLES QU’ELLES SOIENT, expresses ou implicites.
 * Consultez la Licence pour les autorisations et les restrictions linguistiques spécifiques relevant de la Licence.
 *
 * Auteurs:
 *	Pascal Havelange, Pascal.HAVELANGE@ec.europa.eu
 *	Cosmin Popescu,   Adrian.POPESCU@ext.ec.europa.eu
*/

/**
 * Tarîqa 3.0 - Core components
 * @author Pascal Havelange, Cosmin Popescu
 * @copyright © EUROPEAN COMMISSION, 2009-2010
 *
 * @license http://ec.europa.eu/idabc/eupl
 */
namespace tariqa3{
/**
 * Base class for fetching the news results from any engine and from any source
 * @author Cosmin Popescu
 *
 */
class newsAdapter {

    /**
     * The query
     * @var newsQuery
     * @access private
     */
    private $query;

    /**
     * The search configuration (for the translator)
     * @var array
     * @access private
     */
    private $newsConfig = null;

    /**
     * If set to true, then then throw errors. Otherwise, just log the error message
     * @var boolean
     * @access private
     */
    private $throw_errors = false;

    /**
     * Constructs the news adapter
     * @param newsQuery $query The query
     * @param array $config The config array
     * @param boolean $throw_errors If true, throw errors instead of returning a message [OPTIONAL]
     * @access public
     * @return void
     */
    public function __construct($query, $config, $throw_errors = false) {
	$this->query = $query;
	$this->newsConfig = $config['search'];
	$this->throw_errors = $throw_errors;
    }

    /**
     * Checks if an user has access rights
     * @param array $user_groups An array with the user groups
     * @access protected
     * @return boolean
     */
    protected function hasAccess($user_groups) {
	return true;
    }

    /**
     * Translate the query from unified query syntax to engine specific
     * @access protected
     * @return string
     */
    public function translate() {
	$query = $this->query->query;
	if (substr($query, 0, 4) == "UQS:") {
	    $uqs = new \MetaNull\News\unifiedQuerySyntax(substr($query, 4), $this->newsConfig, $this->query->engine_id);
	    $query = $uqs->translate();

	    return $query;
	}
	else if (substr($query, 0, 3) == "OS:") {
	    return substr($query, 3);
	}

	if ($this->throw_errors) {
	    throw new \Exception("Query Syntax error");
	}
	else {
	    error_log("Query Syntax error: " . $this->query->query);
	}

	return null;
    }

    /**
     * Fetches the news results from the SOAP returning an object
     * @access private
     * @return tariqaSearchResultSet
     */
    private function fetchResults() {
	$class_name = 'tariqa3\\search\\' . $this->query->engine_id . 'Engine';
	/* @var $class searchEngine */
	$class = new $class_name($this->newsConfig);
	//$results = $class->listArticles($this->query);
	return $class->listArticles($this->query);
    }

    /**
     * Returns an error to the end user
     * @param boolean $string If set to true, the result will be returned as an XML string, otherwise, the result will
     * be returned as DOMDocument [OPTIONAL]
     * @param boolean $message The error message [OPTIONAL]
     * @access private
     * @return string|DOMDocument
     */
    private function throwError($string, $message = 'Error message N/A') {
	$result = new search\tariqaSearchResultSet();
	$result->query = $this->query->query;
	$result->count = 0;
	$result->engine = "Error";
	$result->source = "Error";

	$document = new search\tariqaDocument();
	$document->title = 'Your search returned an error';
	$document->id = 1;
	$document->date = date('r');
	$document->url = "mailto:RELEX-TARIQA@ec.europa.eu";

	$snippet = <<<html
<div>Your search returned an error. Please contact the <a href="mailto: RELEX-TARIQA@ec.europa.ue"
		>RELEX TARIQA team</a>.</div>
<div>The error message is: </div>
<div><strong>{$message}</strong></div>
<div>You can also refine your search (especially for LexisNexis). If LexisNexis finds more then 3000 results, it will
return an error. </div>
<div>You could also check the syntax of the query. Maybe it is wrong. </div><br/><br/>
<div>Thank you,</div>
<div><b>The RELEX TARIQA Team</b></div>
html;

	$document->summary = $snippet;

	$result->documents = array($document);

	$xml = search\searchEngine::transformToXML($result, $this->newsConfig);

	return ($string?$xml->saveXML():$xml);
    }

    /**
     * Fetches one article
     * @param string $id The document id
     * @param array $newsConfig The news configuration array
     * @param boolean $string If set to true, return the result as a string, else as a DOMDocument [OPTIONAL]
     * @access public
     * @return string|DOMDocument
     */
    public static function fetchArticle($id, $newsConfig, $string = false) {
	//error_log('Check user access');

	$parts = explode(":", $id, 3);
	if (count($parts) != 3) {
	    throw new \Exception("Fetching document error. The id is not in the right format: {$id}");
	}

	$class_name = 'tariqa3\\search\\' . $parts[0] . 'Engine';
	$class = new $class_name($newsConfig);
	/* @var $class searchEngine */
	$document = $class->getArticle($id);
	$xml = search\searchEngine::transformToXML($document, $newsConfig);

	return ($string?$xml->saveXML():$xml);
    }

    /**
     * Fetches the news results
     * @param boolean $string If set to true, the result will be returned as an XML string, otherwise, the result will
     * be returned as DOMDocument [OPTIONAL]
     * @access public
     * @return string|DOMDocument
     */
    public function fetchNews($string = true) {
	$start_date = $this->query->start_date;
	$end_date = $this->query->end_date;
	try {
	    $this->query->query = $this->translate($this->query->query);
	    $c = \MetaNull\News\unifiedQuerySyntax::getDatesCondition($this->query->query);
	    if (array_key_exists('string', $c)) {
		if ($this->query->start_date !== null ||
			$this->query->end_date !== null ||
			$this->query->range !== null ) {
		    throw new \MetaNull\News\unifiedQueryException(
		    'You selected two dates. One from the syntax and one from the form fields. ' .
			    'This is not permited'
		    );
		}
		$this->query->query = $c['query'];
		$range = new timeRange($c['string']);
		$this->query->start_date = $range->getStartDate();
		$this->query->end_date = $range->getEndDate();
	    }
	}
	catch (\Exception $e) {
	    if ($this->throw_errors) {
		\MetaNull\Tools\mDump::dumpVariable($e, true);
		throw new \Exception("Syntax error: " . $e->getMessage());
	    }
	    else {
		error_log("Syntax error: ");
		\MetaNull\Tools\mDump::dumpVariable($e, true);
		return $this->throwError($string, 'Syntax error: ' . $e->getMessage());
	    }
	}

	if ($this->query->query == null && strpos($this->query->engine_id, 'aggregator') === false && $this->query->engine_id != "OxfordAnalytica") {
	    if ($this->throw_errors) {
		throw new \Exception("Syntax error. The query is not defined");
	    }
	    else {
		error_log("Syntax error: " . $this->query->query);
		return $this->throwError($string, 'Empty query.');
	    }
	}

	try {
	    $results = $this->fetchResults();
	    $xml = search\searchEngine::transformToXML($results, $this->newsConfig);
	    //return $results;
	    $this->query->start_date = $start_date;
	    $this->query->end_date = $end_date;
	    return ($string?$xml->saveXML():$xml);
	}
	catch (\Exception $e) {
	    ob_start();

	    var_dump($e);

	    $contents = ob_get_contents();
	    ob_end_clean();
	    error_log($contents);

	    if ($this->throw_errors) {
		throw new \Exception("Engine Exception. Please see the log for details");
	    }
	    else {
		return $this->throwError($string . $e->getMessage(), $e->getMessage());
	    }
	}
    }

    /**
     * Fetches a website content
     * @param boolean $fetch_only_body If true, fetches only the body [OPTIONAL]
     * @access public
     * @return string
     */
    public function fetchWebsite($fetch_only_body = true) {
	try {
	    $page = statics::$tariqaClient->fetchUrl(
		    statics::$tariqaToken, new \WebRequest($this->translate($this->query->query))
	    );

	    $options = array(
		    "drop-proprietary-attributes" => true,
		    "clean" => true,
		    "output-xhtml" => true,
		    "show-body-only" => $fetch_only_body,
	    );
	    $tidy = tidy_parse_string( $page->body, $options, "utf8" );
	    $tidy->cleanRepair();
	    $result = (string)$tidy;
	}
	catch (\Exception $e) {
	    throw new \Exception("Engine Exception. Please the SOAP log for details");
	}

	return $result;
    }
}
}
?>