<?php
	/**
	 * Helper functions & classes to build web pages based on template documents 
	 * @author Pascal (Meta) Havelange (havelangep@gmail.com)
 	 * @copyright &copy; 2007 All rights reserved
 	 * @package MetaNull
 	 */

	/**
	 * Helper class for template document creation
	 * 
	 * The supported syntax is of the form
	 * 
	 * <NAMESPACE:value key="DATA_KEY" [function="FUNCTION_NAME"]>[DEFAULT_VALUE]</NAMESPACE:value>
	 * <NAMESPACE:value key="DATA_KEY" [function="FUNCTION_NAME"]/>
	 * <NAMESPACE:object key="OBJECT_KEY"/>
	 * <NAMESPACE:array key="DATA_KEY">[DATA_BLOCK]</NAMESPACE:array>
	 * <NAMESPACE:until key="FUNCTION_NAME">[DATA_BLOCK]</NAMESPACE:until>
	 * @author Pascal (Meta) Havelange (havelangep@gmail.com)
 	 * @copyright &copy; 2007 All rights reserved
 	 * @package MetaNull
	 */
	class MTemplate {
		
		/**
		 * Template content
		 *
		 * @var string
		 */
		public $template = false;
		
		/**
		 * Array of values
		 *
		 * @var array
		 */
		public $data = false;
		
		/**
		 * Default namespace used by the template
		 *
		 * @var string
		 */
		public $namespace = "mtemplate";
		
		/**
		 * Filepath of the folder where to store the files generated by the cache system
		 *
		 * @access protected
		 * @var string
		 */
		protected $cache_path = false;
		
		/**
		 * Name of the cache file
		 * 
		 * @access protected
		 * @var string
		 */
		protected $cache_file = false;
		
		/**
		 * Validity of the cache file (in seconds)
		 *
		 * @access protected
		 * @var unknown_type
		 */
		protected $cache_validity = 3600;
		
		/**
		 * Constructor
		 *
		 * @return MTemplate
		 */
		function __construct( $namespace = "mtemplate" ) {
			$this->template = false;
			$this->data = array();
			$this->namespace = $namespace;
		}
		
		/**
		 * Transform the loaded template, using the current array of data
		 *
		 * @return string The resulting document
		 */
		function transform() {
			// Cache system enabled?
			if( false !== $this->cache_file ) {
				$fullpath = "";
				if( false != $this->cache_path ) {
					$fullpath .= ereg_replace( "[/\\\\]$", "", $this->cache_path );
					$fullpath = ereg_replace( "[/\\\\]", "/", $fullpath ) . "/";
				}
				$fullpath .= $this->cache_file;
				
				if( !file_exists( $fullpath ) || 
					time() - $this->cache_validity > @filemtime( $fullpath ) ||
					0 == filesize( $fullpath )
				) {
					// File does not exists or has expired... rebuild the template
					$result = $this->apply( $this->template, $this->data );
					$fp = fopen( $fullpath, "wb" );
					if( false === $fp ) {
						// Cache creation failed!
					} else {
						fwrite( $fp, $result );
						fclose( $fp );
						$fp = false;
					}
					return $result;
				} else {
					// Cache is valid, use it "as is"
					return load_file( $fullpath );
				}
			}

			$result = $this->apply( $this->template, $this->data );
			return $result;
		}
		
		/**
		 * Enable or disable the cache system
		 *
		 * @param string $file If FALSE, disable the cache system. Otherwise this must be the filename of the cachefile
		 * @param integer $validity Validity of the cache in seconds
		 * @param string $path Path where to store the cache files (If FALSE, the current folder will be used)
		 */
		function useCache( $file = false, $validity = 3600, $path = false ) {
			$this->cache_file = $file;
			$this->cache_path = $path;
			$this->cache_validity = $validity;
		}

		/**
		 * Transform the passed template into the final "document" using the passed array of data
		 * 
		 * @static 
		 * @access private
		 * @param string $template The template document
		 * @param array $data The array of values
		 * @return string The resulting document
		 */
		private function apply( $template, $data ) {
			$result = $template;
			
			$exp_val = array();
			$exp_idx = array();
			
			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => 3, "attrib" => false );
			$exp_val[] = <<<exp
<{$this->namespace}:(array)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*>(.*)</{$this->namespace}:array[[:blank:]]*>
exp;

			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => 3, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(callback)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*>(.*)</{$this->namespace}:callback[[:blank:]]*>
exp;

			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => 3, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(object)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*>(.*)</{$this->namespace}:object[[:blank:]]*>
exp;

			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => false, "attrib" => true);
			$exp_val[] = <<<exp
@{$this->namespace}:(value)=([a-zA-Z0-9/_-]+)@
exp;

			$exp_idx[] = array( "item" => 1, "key" => 3, "func" => 2, "data" => false, "attrib" => true);
			$exp_val[] = <<<exp
@{$this->namespace}:(value)=([a-zA-Z0-9/_-]+)\\(([a-zA-Z0-9/_-]+)\\)@
exp;

$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => false, "attrib" => false );
			$exp_val[] = <<<exp
@{$this->namespace}:(rawvalue)=([a-zA-Z0-9/_-]+)@
exp;

			$exp_idx[] = array( "item" => 1, "key" => 3, "func" => 2, "data" => false, "attrib" => false );
			$exp_val[] = <<<exp
@{$this->namespace}:(rawvalue)=([a-zA-Z0-9/_-]+)\\(([a-zA-Z0-9/_-]+)\\)@
exp;
			
			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => 3, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(value)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*>(.*)</{$this->namespace}:value[[:blank:]]*>
exp;
			
			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => 3, "data" => 4, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(value)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*function[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*>(.*)</{$this->namespace}:value[[:blank:]]*>
exp;
			$exp_idx[] = array( "item" => 1, "key" => 3, "func" => 2, "data" => 4, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(value)[[:blank:]]*function[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*>(.*)</{$this->namespace}:value[[:blank:]]*>
exp;

			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => false, "data" => false, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(value)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*/>
exp;

			$exp_idx[] = array( "item" => 1, "key" => 2, "func" => 3, "data" => false, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(value)[[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*function[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*/>
exp;

			$exp_idx[] = array( "item" => 1, "key" => 3, "func" => 2, "data" => false, "attrib" => false);
			$exp_val[] = <<<exp
<{$this->namespace}:(value)[[:blank:]]*function[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_\\-]+)["'][[:blank:]]*key[[:blank:]]*=[[:blank:]]*["']([a-zA-Z0-9/_-]+)["'][[:blank:]]*/>
exp;

			foreach( $exp_val as $idx => $exp ) {
				
				$regs = array();
				while( ereg( $exp, $result, $regs ) ) {
					
					/**
					 * array :=> apply the template block for each element of the passed array. The elements of the array are expected to be array of values to use in the template transformation.
					 */
					if( "array" == $regs[ $exp_idx[$idx]["item"]] ) {
						
						// Test if no other callback elements are nested
						// Recherche la première ouverture/fermeture d'un tag identique
						$first_opening = strpos( $regs[ $exp_idx[$idx]["data"] ], "<{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}" );
						$first_closing = strpos( $regs[ $exp_idx[$idx]["data"] ], "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>" );
						while( false !== $first_opening && false !== $first_closing ) {
							if( $first_closing < $first_opening ) {
								// Restricts the expression
								$opening_tab = strpos( $regs[0], $regs[ $exp_idx[$idx]["data"] ] );
								$regs[0] = substr( $regs[0], 0, $opening_tab + $first_closing ) . "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>";
								
								// Restricts the data
								$regs[ $exp_idx[$idx]["data"] ] = substr( $regs[ $exp_idx[$idx]["data"] ], 0, $first_closing );
								break;
							}
							// Si la première fermeture arrive APRES la première ouverture
							// --> Il s'agit d'un élément imbriqué, d'un niveau inférieur
							// ==> Continue la recherche plus loin
							// Recherche la première ouverture/fermeture d'un tag identique suivante
							$first_closing = strpos( $regs[ $exp_idx[$idx]["data"] ], "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>", $first_closing + 1 );
							$first_opening = strpos( $regs[ $exp_idx[$idx]["data"] ], "<{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}", $first_closing + 1 );
						}

						$block = $regs[ $exp_idx[$idx]["data"] ];
						if( !isset( $data[ $regs[ $exp_idx[$idx]["key"] ] ] )) {
							$result = str_replace( $regs[0], "", $result );
							continue;
						} 
						
						if(!is_array( $data[ $regs[ $exp_idx[$idx]["key"] ] ] )) {
							$result = str_replace( $regs[0], "", $result );
							trigger_error( "mtemplate::apply({$regs[0]}):\n-->\$data[{$regs[ $exp_idx[$idx]["key"] ]}] Is not an Array!" );
							continue;
						}
						
						$cells = "";
						foreach( $data[ $regs[ $exp_idx[$idx]["key"] ] ] as $data_cell ) {
							$cells .= $this->apply( $block, array_merge( $data, $data_cell ));
						}
						$result = str_replace( $regs[0], $cells, $result );
						
					} else 
					
					/**
					 * callback :=> apply the template block until the passed function returns false. This function is expected to return an array of values to use in the template transformation
					 */
					if( "callback" == $regs[ $exp_idx[$idx]["item"]] ) {
						
						// Test if no other callback elements are nested
						// Recherche la première ouverture/fermeture d'un tag identique
						$first_opening = strpos( $regs[ $exp_idx[$idx]["data"] ], "<{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}" );
						$first_closing = strpos( $regs[ $exp_idx[$idx]["data"] ], "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>" );
						while( false !== $first_opening && false !== $first_closing ) {
							if( $first_closing < $first_opening ) {
								// Restricts the expression
								$opening_tab = strpos( $regs[0], $regs[ $exp_idx[$idx]["data"] ] );
								$regs[0] = substr( $regs[0], 0, $opening_tab + $first_closing ) . "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>";
								
								// Restricts the data
								$regs[ $exp_idx[$idx]["data"] ] = substr( $regs[ $exp_idx[$idx]["data"] ], 0, $first_closing );
								break;
							}
							// Si la première fermeture arrive APRES la première ouverture
							// --> Il s'agit d'un élément imbriqué, d'un niveau inférieur
							// ==> Continue la recherche plus loin
							// Recherche la première ouverture/fermeture d'un tag identique suivante
							$first_closing = strpos( $regs[ $exp_idx[$idx]["data"] ], "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>", $first_closing + 1 );
							$first_opening = strpos( $regs[ $exp_idx[$idx]["data"] ], "<{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}", $first_closing + 1 );
						}
						
						$block = $regs[ $exp_idx[$idx]["data"] ];
						if( !isset( $data[ $regs[ $exp_idx[$idx]["key"] ] ] )) {
							$result = str_replace( $regs[0], "", $result );
							continue;
						} 
						
						if( !function_exists( $data[ $regs[ $exp_idx[$idx]["key"] ] ] )){
							$result = str_replace( $regs[0], "", $result );
							trigger_error( "mtemplate::apply({$regs[0]}): \$data[{$regs[ $exp_idx[$idx]["key"] ]}] Is not a valid callback!" );
							continue;
						}
						
						$cells = "";
						$data_cell = false;
						while( false !== ( $data_cell = call_user_func( $data[ $regs[ $exp_idx[$idx]["key"] ] ] ) )) {
							$cells .= $this->apply( $block, array_merge( $data, $data_cell ));
						}
						$result = str_replace( $regs[0], $cells, $result );
						
					} else 
					
					/**
					 * object :=> apply the template block using the passed object (the object must extends MTemplateObject)
					 */
					if( "object" == $regs[ $exp_idx[$idx]["item"]] ) {

						// Test if no other callback elements are nested
						// Recherche la première ouverture/fermeture d'un tag identique
						$first_opening = strpos( $regs[ $exp_idx[$idx]["data"] ], "<{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}" );
						$first_closing = strpos( $regs[ $exp_idx[$idx]["data"] ], "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>" );
						while( false !== $first_opening && false !== $first_closing ) {
							if( $first_closing < $first_opening ) {
								// Restricts the expression
								$opening_tab = strpos( $regs[0], $regs[ $exp_idx[$idx]["data"] ] );
								$regs[0] = substr( $regs[0], 0, $opening_tab + $first_closing ) . "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>";
								
								// Restricts the data
								$regs[ $exp_idx[$idx]["data"] ] = substr( $regs[ $exp_idx[$idx]["data"] ], 0, $first_closing );
								break;
							}
							// Si la première fermeture arrive APRES la première ouverture
							// --> Il s'agit d'un élément imbriqué, d'un niveau inférieur
							// ==> Continue la recherche plus loin
							// Recherche la première ouverture/fermeture d'un tag identique suivante
							$first_closing = strpos( $regs[ $exp_idx[$idx]["data"] ], "</{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}>", $first_closing + 1 );
							$first_opening = strpos( $regs[ $exp_idx[$idx]["data"] ], "<{$this->namespace}:{$regs[ $exp_idx[$idx]["item"]]}", $first_closing + 1 );
						}
						
						$block = $regs[ $exp_idx[$idx]["data"] ];
						if( !isset( $data[ $regs[ $exp_idx[$idx]["key"] ] ] )) {
							$result = str_replace( $regs[0], "", $result );
							continue;
						} 
						
						if( !is_object( $data[ $regs[ $exp_idx[$idx]["key"] ] ] ) || !is_a( $data[ $regs[ $exp_idx[$idx]["key"] ] ], "MTemplateObject" )){
							$result = str_replace( $regs[0], "", $result );
							trigger_error( "mtemplate::apply({$regs[0]}):\n-->\$data[{$regs[ $exp_idx[$idx]["key"] ]}] Is not a valid MTemplateObject instance!" );
							continue;
						}
						
						$cells = "";
						$data_cell = false;
						if( false === $data[ $regs[ $exp_idx[$idx]["key"] ] ]->initialize( $data ) ) {
							$result = str_replace( $regs[0], $data[ $regs[ $exp_idx[$idx]["key"] ] ]->error(), $result );
							trigger_error( "mtemplate::apply({$regs[0]}):\n-->\$data[{$regs[ $exp_idx[$idx]["key"] ]}] MTemplateObject initialisation failed!\n" . $data[ $regs[ $exp_idx[$idx]["key"] ] ]->error() );
							$data[ $regs[ $exp_idx[$idx]["key"] ] ]->destroy();
							continue;
						}
						while( false !== ( $data_cell = $data[ $regs[ $exp_idx[$idx]["key"] ] ]->next() )) {
							$cells .= $this->apply( $block, array_merge( $data, $data_cell ));
						}
						$result = str_replace( $regs[0], $cells, $result );
						$data[ $regs[ $exp_idx[$idx]["key"] ] ]->destroy();
					} else 
					
					/**
					 * value :=> replace the template instruction by the requested value. Possibly first apply to this value the predefined function.
					 */
					{

						if( isset( $data[ $regs[ $exp_idx[$idx]["key"] ] ] )) {
							// Get the value from the data array
							$rpl_value = $data[ $regs[ $exp_idx[$idx]["key"] ] ];
							// If "function" is set, first execute the callback
							if( false !== $exp_idx[$idx]["func"] ) {
								$func_name = $regs[ $exp_idx[$idx]["func"]];
								if( function_exists( $func_name ) ) {
									$rpl_value = call_user_func( $func_name, $rpl_value );
								} else {
									trigger_error( "mtemplate::apply({$regs[0]}): {$func_name}() Is not a function!" );
								}
							} 
							
							// If "attribute" is set, escape the string (removes quotes and double quotes)
							if( false !== $exp_idx[$idx]["attrib"] ) {
								$rpl_value = str_replace( array("\"", "'"), array("&#34;", "&#39;"), $rpl_value );
							}
							
							// Value found in data
							$result = str_replace( $regs[0], $rpl_value, $result );
						} else {
							if( false === $exp_idx[$idx]["data"] ) {
								// Value not found in data, and no default value => erase
								$result = str_replace( $regs[0], "", $result );
							} else {
								// Value not found in data, and default value set => use default
								$result = str_replace( $regs[0], $regs[ $exp_idx[$idx]["key"] ], $result );
							}
						}
					} 
						
				}
				
			}
			
			return $result;

		}
		

		/**
		 * Load the template from a file
		 *
		 * @param string $path
		 */
		function loadTemplate( $path ) {
			$this->template = load_file( $path );
		}
	}
	
	/**
	 *	Reads an entire file and returns it's content
	 *
	 * @param string $filename Full path of the file
	 * @param integer $block_size Default size of the block for read operation
	 * @return string
	 * @author Pascal (Meta) Havelange (havelangep@gmail.com)
 	 * @copyright &copy; 2007 All rights reserved
	 */
	function load_file( $filename, $block_size = 8192 ) {
		
		if( function_exists( "file_get_contents" )) {
			return file_get_contents( $filename, FALSE );
		} else {
			$fp = @fopen( $filename, "rb" );
			if( false == $fp ) {
				return false;
			}
			$data = "";
			while( !feof( $fp )) {
				$data .= fread( $fp, $block_size );
			}
			fclose( $fp );
			return $data;
		}
	}
	
	/**
	 * Interface for objects used by the MTemplate class
	 * Evey mtemplate:object key must refer to a class extending MTemplateObject
	 * @author Pascal (Meta) Havelange (havelangep@gmail.com)
 	 * @copyright &copy; 2007 All rights reserved
 	 * @package MetaNull
	 */
	abstract class MTemplateObject {
		abstract public function initialize( $data = null );
		abstract public function next();
		abstract public function destroy();
		abstract public function error();
	}
	
	/**
	 * Specialize MTemplateObject class to help handle MySQL queries
	 * 
	 * Example:
	 * -------- 
	 * - Template: (file: persons.html)
	 * <html><body>
	 * <table>
	 * <m:object key="my_query">
	 * 	<tr><td><m:value key="last_name" /></td>
	 * 		<td><m:value key="first_name" /></td>
	 * </tr>
	 * </m:object>
	 * </table>
	 * </body></html>
	 * 
	 * - PHP: (file: persons.php)
	 * <?php
	 * $my_query = new MTemplateObject_mysql();
	 * $my_query->user = "guest";
	 * $my_query->pass = "";
	 * $my_query->host = "localhost";
	 * $my_query->name = "test";
	 * $my_query->encoding = "utf8";
	 * $my_query->query = "select first_name, last_name from persons";
	 * $page = new MTemplate( "m" );
	 * $page->loadTemplate( "persons.html" );
	 * $page->data["my_query"] = $my_query;
	 * $page->useCache( FALSE );
	 * echo $page->transform();
	 * ?>
	 * 
	 * @author Pascal (Meta) Havelange (havelangep@gmail.com)
 	 * @copyright &copy; 2007 All rights reserved
 	 * @package MetaNull
	 */
	class MTemplateObject_mysql extends MTemplateObject {
		
		protected $db = false;
		protected $rc = false;
		protected $query = false;
		
		public $host = "localhost";
		public $name = "test";
		public $user = "guest";
		public $encoding = "utf-8";
		public $pass = null;
		
		
		function __construct() {
			parent::__construct();
		}
		
		function initialize( $data = null ) {
			
			$data;	// $data is not used in this specific class
			
			$this->db = @mysql_connect( $this->host, $this->user, $this->pass );
			if( false === $this->db ) {
				return false;
			}
			if( false === @mysql_select_db( $this->name, $this->db )) {
				return false;
			}
			if( false === @mysql_query( "set names {$this->encoding}", $this->db )) {
				return false;
			}
			
/*			$query = $this->query;
			while( false !== ereg( "@@([a-zA-Z0-9/\\_-]+)@@", $query, $regs )) {
				$query = str_replace( $regs[0], $data[$regs[1]], $query );
			}
			error_log( $query );
*/
			$this->rc = @mysql_query( $this->query, $this->db );
			if( false === $this->rc ) {
				@mysql_close( $this->db );
				$this->db = false;
				return false;
			}
			return true;
		}
		
		function next() {
			if( false === $this->rc ) {
				return false;
			}
			$row = @mysql_fetch_array( $this->rc, MYSQL_ASSOC );
			return $row;
		}
		
		function destroy() {
			if( false !== $this->db ) {
				@mysql_close( $this->db );
				$this->db = false;
			}
		}
		
		function error() {
			if( false === $this->db ) {
				return "Could not connect to MySQL server on {$this->host}: \n" . mysql_error();
			}
			$errc = @mysql_errno( $this->db );
			$errm = @mysql_error( $this->db );
			return <<<error
MySQL Error #{$errc}:
{$errm}

In SQL query:
{$this->query}

error;
		}
		
	}

	// Some general purpose callbacks to be used in the template function
	/**
	 * Escapes HTML special characters
	 * @param string $data
	 * @return string
	 */
	function cb_html_escape( $data ) {
		return htmlspecialchars( $data, ENT_COMPAT, "UTF-8" );
	}
	
	/**
	 * Escapes a string so that it can be used as an attribute value in an HTML document
	 * @param string $data
	 * @return string
	 */
	function cb_html_attribute_value( $data ) {
		return str_replace("\r\n"," ", htmlspecialchars( $data, ENT_QUOTES, "UTF-8"));
	}
	
	/**
	 * Cleanup HTML data to make it fits strictly with XHTML format
	 * @param string $data
	 * @return string
	 */
	function cb_html_tidy( $data ) {
		$config = array(
			"output-xhtml" => true,
			"show-body-only" => true,
			"doctype" => "strict",
		
			"indent" => true,
			"drop-proprietary-attributes" => true,
			"lower-literals" => true,
			"drop-font-tags" => true,
			"quote-ampersand" => true,
			"wrap" => 120,
		);
		$tidy = new tidy;
		$tidy->parseString( $data, $config, "utf8" );
		$tidy->cleanRepair();
		return $tidy;
	}
	
	
	/**
	 * Convert CR & LF to HTML paragraphs
	 * @param string $data
	 * @return string
	 */
	function cb_crlf_to_br( $data ) {
		$str = ereg_replace( "\r\n", "\n", $data );
		$str = ereg_replace( "<br/>[\n]?", "\n", $str );
		$str_paragraphs = split( "\n\n", $str );
		foreach( $str_paragraphs as $pidx => $paragraph ) {
			$str_lines = split( "\n", $paragraph );
			foreach( $str_lines as $lidx => $line ) {
				$str_lines[$lidx] = htmlspecialchars( $line, ENT_COMPAT, "UTF-8" );
			}
			$str_paragraphs[$pidx] = implode( "<br/>\n", $str_lines );
		}
		if( count( $str_paragraphs ) > 1 ) {
			$str_paragraphs = array_filter( array_map("trim", $str_paragraphs));
			$str = "<p>".implode("</p>\n<p>", $str_paragraphs )."</p>\n";			
		} else {
			$str = $str_paragraphs[0];
		}
		return $str;
		
	}
	
?>